#  Copyright (C) 2024 Yang. - All Rights Reserved
"""
# ==========================================
# Copyright 2024 Yang
# ararpy - test.py
# ==========================================
#
#
#
"""
import ararpy as ap
import numpy as np
import pdf_maker as pm
import os

import matplotlib

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


def export_pdf(each_line, X, Y, x, y, Tc_list, sample_name):

    example_dir = r"C:\Users\Young\OneDrive\00-Projects\【2】个人项目\2022-05论文课题\初稿\封闭温度计算"

    # ------ 将以下五个样品的年龄谱图组合到一起 -------
    arr_files = [
        os.path.join(example_dir, r'22WHA0433.arr'),
        os.path.join(example_dir, r'20WHA0103.age'),
    ]
    colors = ['#1f3c40', '#e35000', '#e1ae0f', '#3d8ebf']
    series = []

    # ------ 构建数据 -------
    for index, file in enumerate(arr_files):
        smp = ap.from_arr(file_path=file) if file.endswith('.arr') else ap.from_age(file_path=file)
        age = smp.ApparentAgeValues[2:4]
        ar = smp.DegasValues[20]
        data = ap.calc.spectra.get_data(*age, ar, cumulative=False)
        series.append({
            'type': 'series.line', 'id': f'line{index * 2 + 0}', 'name': f'line{index * 2 + 0}', 'color': colors[index],
            'data': np.transpose([data[0], data[1]]).tolist(), 'line_caps': 'square',
        })
        series.append({
            'type': 'series.line', 'id': f'line{index * 2 + 1}', 'name': f'line{index * 2 + 1}', 'color': colors[index],
            'data': np.transpose([data[0], data[2]]).tolist(), 'line_caps': 'square',
        })
        series.append({
            'type': 'text', 'id': f'text{index * 2 + 0}', 'name': f'text{index * 2 + 0}', 'color': colors[index],
            'text': f'{smp.name()}<r>{round(smp.Info.results.age_plateau[0]["age"], 2)}',
            'size': 10, 'data': [[index * 15 + 5, 23]],
        })
    data = {
        "data": [
            {
                'xAxis': [{'extent': [0, 100], 'interval': [0, 20, 40, 60, 80, 100],
                           'title': 'Cumulative <sup>39</sup>Ar Released (%)', 'nameLocation': 'middle', }],
                'yAxis': [{'extent': [0, 250], 'interval': [0, 50, 100, 150, 200, 250],
                           'title': 'Apparent Age (Ma)', 'nameLocation': 'middle', }],
                'series': series
            },
            {
                'xAxis': [{'extent': [0, 100], 'interval': [0, 20, 40, 60, 80, 100],
                           'title': 'Cumulative <sup>39</sup>Ar Released (%)', 'nameLocation': 'middle', }],
                'yAxis': [{'extent': [0, 250], 'interval': [0, 50, 100, 150, 200, 250],
                           'title': 'Apparent Age (Ma)', 'nameLocation': 'middle', }],
                'series': series
            }
        ],
        "file_name": f"{sample_name}",
        "plot_names": [f"{sample_name}"],
    }

    # write pdf
    file = pm.NewPDF(filepath=os.path.join(example_dir, f"{data['file_name']}.pdf"))
    for index, each in enumerate(data['data']):
        # rich text tags should follow this priority: color > script > break
        file.text(page=index, x=50, y=780, line_space=1.2, size=12, base=0, h_align="left",
                  text=f"The PDF can be edited with Adobe Acrobat, Illustrator and CorelDRAW")
        cv = ap.smp.export.export_chart_to_pdf(each)
        file.canvas(page=index, base=0, margin_top=5, canvas=cv, unit="cm", h_align="middle")
        if index + 1 < len(data['data']):
            file.add_page()

    # save pdf
    file.save()


def read_sample(arr_file_path):
    # 读取样品信息
    if not arr_file_path.endswith('.arr'):
        for root, dirs, files in os.walk(arr_file_path):
            for file in files:
                if file.endswith('.arr'):
                    arr_file_path = os.path.join(arr_file_path, file)
                    break
    print(f"arr file: {arr_file_path}")
    sample = ap.from_arr(file_path=arr_file_path)
    name = sample.name()
    sequence = sample.sequence()
    nsteps = sequence.size
    te = np.array(sample.TotalParam[124], dtype=np.float64)
    ti = (np.array(sample.TotalParam[123], dtype=np.float64) / 60).round(2)  # time in minute
    nindex = {"40": 24, "39": 20, "38": 10, "37": 8, "36": 0}
    argon = "39"
    if argon in list(nindex.keys()):
        ar = np.array(sample.DegasValues[nindex[argon]], dtype=np.float64)  # 20-21 Argon
        sar = np.array(sample.DegasValues[nindex[argon] + 1], dtype=np.float64)
    elif argon == 'total':
        all_ar = np.array(sample.CorrectedValues, dtype=np.float64)  # 20-21 Argon
        ar, sar = ap.calc.arr.add(*all_ar.reshape(5, 2, len(all_ar[0])))
        ar = np.array(ar)
        sar = np.array(sar)
    else:
        raise KeyError
    age = np.array(sample.ApparentAgeValues[2], dtype=np.float64)  # 2-3 age
    sage = np.array(sample.ApparentAgeValues[3], dtype=np.float64)
    f = np.cumsum(ar) / ar.sum()

    # 组合data
    dr2 = [1 for i in range(nsteps)]
    ln_dr2 = [1 for i in range(nsteps)]
    wt = [1 for i in range(nsteps)]
    data = np.array([sequence.value, te, ti, age, sage, ar, sar, f, dr2, ln_dr2, wt]).tolist()
    data.insert(0, (np.where(np.array(data[3]) > 0, True, False) & np.isfinite(data[3])).tolist())
    data.insert(1, [1 for i in range(nsteps)])
    for row in ap.calc.arr.transpose(data):
        print(row)
    return data, name


def calculate_dr2(f, ti, ar, sar, use_ln=True, logdr2_method="plane"):
    try:
        if str(logdr2_method).lower().startswith('plane'.lower()):
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_plane(f, ti, ar=ar, sar=sar, ln=use_ln)
        elif str(logdr2_method).lower() == 'yang':
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_yang(f, ti, ar=ar, sar=sar, ln=use_ln)
        elif str(logdr2_method).lower().startswith('sphere'.lower()):
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_sphere(f, ti, ar=ar, sar=sar, ln=use_ln)
        elif str(logdr2_method).lower().startswith('Thern'.lower()):
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_thern(f, ti, ar=ar, sar=sar, ln=use_ln)
        elif str(logdr2_method).lower().startswith('cylinder'.lower()):
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_cylinder(f, ti, ar=ar, sar=sar, ln=use_ln)
        elif str(logdr2_method).lower().startswith('cube'.lower()):
            dr2, ln_dr2, wt = ap.smp.diffusion_funcs.dr2_cube(f, ti, ar=ar, sar=sar, ln=use_ln)
        else:
            raise KeyError(f"Geometric model not found: {str(logdr2_method).lower()}")
    except (Exception, BaseException) as e:
        raise ValueError

    return dr2, ln_dr2, wt


def calculate_tc(arr_file_path, logdr2_method='plane', A=55, cooling_rate=10, radius=100, use_ln=True, index=None):
    base = np.e if use_ln else 10
    # [A, cooling_rate, radius] = [55, 10, 100]  # 55, 10°C/Ma, 100µm

    # 读取样品信息
    [_, _, seq_value, te, ti, age, sage, ar, sar, f, dr2, ln_dr2, wt], name = read_sample(arr_file_path=arr_file_path)
    dr2, ln_dr2, wt = calculate_dr2(f, ti, ar, sar, use_ln=use_ln, logdr2_method=logdr2_method)
    te = [i + 273.15 for i in te]

    @np.vectorize
    def get_da2_e_Tc(b, m):
        k1 = base ** b * ap.thermo.basic.SEC2YEAR  # k1: da2
        k2 = -10 * m * ap.thermo.basic.GAS_CONSTANT * np.log(base)  # activation energy, kJ
        try:
            # Closure temperature
            k3, _ = ap.thermo.basic.get_tc(da2=k1, sda2=0, E=k2 * 1000, sE=0, pho=0,
                                           cooling_rate=cooling_rate, A=A)
        except ValueError as e:
            # print(e.args)
            k3 = 999
        return k1, k2, k3  # da2, E, Tc

    index = np.s_[index if index is not None else list(range(len(te)))]
    each_line = [np.nan for i in range(17)]  # [b, sb, a, sa, ..., energy, se, tc, stc]
    temp_err = 5
    X, Y, wtX, wtY = 10000 / np.array(te), np.array(ln_dr2), 10000 * temp_err / np.array(te) ** 2, np.array(wt)
    x, y, wtx, wty = X[index,], Y[index,], wtX[index,], wtY[index]
    Tc_list = []

    if len(x) > 0:

        for cooling_rate in np.linspace(start=0, stop=100, num=100):
            # Arrhenius line regression
            # each_line[0:6] = ap.thermo.basic.fit(x, y, wtx, wty)  # intercept, slop, sa, sb, chi2, q
            # b (intercept), sb, a (slope), sa, mswd, dF, Di, k, r2, chi_square, p_value, avg_err_s, cov
            each_line[0:13] = ap.calc.regression.york2(x, wtx, y, wty, ri=np.zeros(len(x)))
            each_line[1] = each_line[1] * 1  # 1 sigma
            each_line[3] = each_line[3] * 1  # 1 sigma

            # monte carlo simulation with 4000 trials
            cov_matrix = np.array([[each_line[1] ** 2, each_line[12]], [each_line[12], each_line[3] ** 2]])
            mean_vector = np.array([each_line[0], each_line[2]])
            random_numbers = np.random.multivariate_normal(mean_vector, cov_matrix, 4000)
            res, cov = ap.calc.basic.monte_carlo(get_da2_e_Tc, random_numbers, confidence_level=0.95)
            da2, E, Tc = res[0:3, 0]
            # sda2, sE, sTc = np.diff(res[0:3, [1, 2]], axis=1).flatten() / 2
            sda2, sE, sTc = 2 * cov[0, 0] ** .5, 2 * cov[1, 1] ** .5, 2 * cov[2, 2] ** .5  # 95%

            each_line[13:15] = [E, sE]
            each_line[15:17] = [Tc, sTc]

            Tc_list.append([cooling_rate, Tc, sTc])

    Tc_list = np.transpose(Tc_list)

    return each_line, X, Y, x, y, Tc_list, name


def plot_Tc(file_path, index=None, radius=100, use_ln=True):
    if index is None:
        index = [1, 2, 3, 4, 5]
    logdr2_methods = ['plane', 'cylinder', 'sphere']
    As = [8.7, 27, 55]

    x1_extent = [10, 0]
    y1_extent = [-5, -10]

    x2_extent = [0, 100]
    y2_extent = [0, 450]

    fig, axs = plt.subplots(len(logdr2_methods), 2, figsize=(12, 12))
    for i in range(len(logdr2_methods)):
        logdr2_method = logdr2_methods[i]
        A = As[i]
        each_line, X, Y, x, y, Tc_list, name = calculate_tc(file_path, logdr2_method=logdr2_method, A=A, radius=radius,
                                                      use_ln=use_ln, index=index, cooling_rate=10)
        print(each_line)
        [E, sE] = each_line[13:15]
        axs[i, 0].scatter(X, Y, c='white', edgecolors='black')
        axs[i, 0].scatter(x, y, c='red')
        axs[i, 0].plot([min(x), max(x)], [i * each_line[2] + each_line[0] for i in [min(x), max(x)]], c='blue')
        axs[i, 0].set_title(f'{name} - Arrhenius Plot - {logdr2_method}')
        axs[i, 0].set_xlabel(f'10000 / T')
        axs[i, 0].set_ylabel(f'ln(D/r2)' if use_ln else f'log(D/r2)')
        axs[i, 0].text(axs[0, 0].get_xlim()[0], np.average(axs[0, 0].get_ylim()),
                       f'intercept = {each_line[0]:.6f} ± {each_line[1]:.6f}\n'
                       f'slope = {each_line[2]:.6f} ± {each_line[3]:.6f}\n'
                       f'E = {E / 4.184:.6f} ± {2 * sE / 4.184:.6f} kcal/mol (2sigma)\n'
                       f'log(D0/r2) = {each_line[0]:.6f} ± {2 * each_line[1]:.6f} /s (2sigma)\n'
                       f'A = {A}\n'
                       f'radius = {radius} µm')

        for s in Tc_list[2]:
            axs[i, 1].plot([Tc_list[0], Tc_list[0]], [Tc_list[1]-2*s, Tc_list[1]+2*s], c="#f0f0f0")
            axs[i, 1].plot([Tc_list[0], Tc_list[0]], [Tc_list[1]-s, Tc_list[1]+s], c="#8e8e8e")

        axs[i, 1].plot(Tc_list[0], Tc_list[1], c="black")
        axs[i, 1].set_ylim(*[y2_extent])
        axs[i, 1].set_xlim(*[x2_extent])
        axs[i, 1].set_xlabel(f'Cooling Rate')
        axs[i, 1].set_ylabel(f'Closure Temperature')

        x1_extent = [min(axs[i, 0].get_xlim()[0], x1_extent[0]), max(axs[i, 0].get_xlim()[1], x1_extent[1])]
        y1_extent = [min(axs[i, 0].get_ylim()[0], y1_extent[0]), max(axs[i, 0].get_ylim()[1], y1_extent[1])]


    for i in range(len(logdr2_methods)):
        axs[i, 0].set_xlim(*[x1_extent])
        axs[i, 0].set_ylim(*[y1_extent])


    print(Tc_list[0])
    fig.tight_layout()
    plt.show()


def export_ar_relase_curve():

    plot_data = {
        "data": [],
        "file_name": f"-Tc",
        "plot_names": [f"-Tc"],
    }

    plot_data['data'].append({
        'xAxis': [{
            'extent': [400, 1500], 'interval': [400, 600, 800, 1000, 1200, 1400, 1600],
            'title': f'Temperature [°C]', 'nameLocation': 'middle',
            'show_frame': True, 'label_size': 10, 'title_size': 10,
        }],
        'yAxis': [{
            'extent': [0, 100], 'interval': [0, 20, 40, 60, 80, 100],
            # 'extent': [-25, 0], 'interval': [-30, -25, -20, -15, -10, -5, 0],
            'title': f'Cumulative argon released [%]', 'nameLocation': 'middle',
            'show_frame': True, 'label_size': 10, 'title_size': 10,
        }],
        'series': [
        ]
    })

    plot_data['data'].append({
        'xAxis': [{
            'extent': [0, 100], 'interval': [0, 20, 40, 60, 80, 100],
            'title': 'Cooling Rate [°C/Ma]', 'nameLocation': 'middle', 'show_frame': True,
            'label_size': 10, 'title_size': 10,
        }],
        'yAxis': [{
            'extent': [0, 850], 'interval': [0, 100, 200, 300, 400, 500, 600, 700, 800],
            'title': 'Closure Temperature [°C]', 'nameLocation': 'middle', 'show_frame': True,
            'label_size': 10, 'title_size': 10,
        }],
        'series': []
    })

    colors = {"bt": 'black', 'ms': 'red', 'tour': 'grey', 'amp': 'blue', 'pl': '#a4df82', 'kfs': '#a43c82'}
    sample_types = [
        r"bt",
        r"ms",
        r"bt",
        r"bt",
        r"ms",
        r"ms",
        r"bt",
        r"ms",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"kfs",
        r"tour",
        r"tour",
        r"tour",
        r"tour",
        r"tour",
        r"tour",
        r"amp",
        r"amp",
        r"amp",
        r"amp",
        r"amp",
    ]
    files = [
        r"D:\DjangoProjects\webarar\static\download\20241121_24FY01a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241123_24FY02a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241125_24FY03a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241129_24FY06a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241130_24FY07a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20241201_24FY08a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241202_24FY09a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241203_24FY10a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240630_24FY49a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240705_24FY50a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240710_24FY51a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240714_24FY52.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240728_24FY55a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240801_24FY61.arr",
        r"D:\DjangoProjects\webarar\private\upload\20241016_24FY66a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240821_24FY67a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240924_24FY70a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20240823_24FY71a.arr",
        r"D:\DjangoProjects\webarar\private\upload\20241212_24FY80a.arr",
        r"C:\Users\Young\OneDrive\00-Projects\【2】个人项目\2022-05论文课题\初稿\[01]ArAr\Gaowu\20241115_24FY81a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241210_24FY82a.arr",
        r"D:\DjangoProjects\webarar\static\download\20241019_24FY83a.arr",
        r"D:\DjangoProjects\webarar\private\upload\Table_M - BH04 diffusion experiment.arr",
        r"D:\DjangoProjects\webarar\private\mdd\MA08-1\Table_K - MA08 diffusion experiment.arr",
        r"D:\DjangoProjects\webarar\private\mdd\126KD57\126KD57-Dinh2023.arr",
        r"D:\DjangoProjects\webarar\private\mdd\Growdon2013\CT0714-Growdon2013.arr",
        r"D:\DjangoProjects\webarar\private\mdd\Growdon2013\CT0705A-Growdon2013.arr",
        r"D:\DjangoProjects\webarar\private\mdd\Growdon2013\CT0712-Growdon2013.arr",
        r"D:\DjangoProjects\webarar\private\mdd\Growdon2013\CT0726-Growdon2013.arr",
    ]
    for arr_file, sample_type in zip(files, sample_types):
        print(f"{sample_type = }, {arr_file = }")
        sample = ap.from_arr(file_path=arr_file)
        te = np.array(sample.TotalParam[124], dtype=np.float64)
        ar = np.array(sample.DegasValues[20], dtype=np.float64)  # 20-21 Argon 39
        f = np.cumsum(ar) / ar.sum() * 100

        plot_data['data'][-2]['series'].append(
            {
                'type': 'series.line', 'id': f'scatter_line', 'name': f'scatters_line',
                'color': colors[sample_type], 'line_width': 1,
                'data': np.transpose([te, f]).tolist(), 'line_caps': 'none',
            },
        )

    params_list = {
        "page_size": 'a4', "ppi": 72, "width": 9.5, "height": 7,
        "pt_width": 0.8, "pt_height": 0.8, "pt_left": 0.16, "pt_bottom": 0.18,
        "offset_top": 0, "offset_right": 0, "offset_bottom": 25, "offset_left": 35,
        "plot_together": False, "show_frame": True,
    }

    filename = f"释放曲线"
    filepath = os.path.join(r"C:\Users\Young\Downloads", f"{filename}.pdf")
    cvs = [[ap.smp.export.get_cv_from_dict(plot, **params_list) for plot in plot_data['data']]]
    filepath = ap.smp.export.export_chart_to_pdf(cvs, filename, filepath)



if __name__ == "__main__":
    file_path = r"D:\DjangoProjects\webarar\private\mdd\Y01\20241121_24FY01a-removeHT.arr"
    index = [1, 2, 3, 4, 5, 6, 7, 8]

    # file_path = r"D:\DjangoProjects\webarar\private\mdd\MA08-2\Table_L - MA08 diffusion experiment.arr"
    # index = [0, 1, 2, 3, 4, 5]
    # # plot_Tc(file_path, index=index)
    #
    # file_path = r"D:\DjangoProjects\webarar\private\mdd\MA08-1\Table_K - MA08 diffusion experiment.arr"
    # index = [4, 5, 6, 7, 8]
    #
    # file_path = r"D:\DjangoProjects\webarar\private\mdd\CR99NW\CR99NW Shi 2020.arr"
    # index = list(range(10))


    # file_path = r"D:\DjangoProjects\webarar\private\mdd\Y07"
    # index = list(range(1, 9))
    #
    # plot_Tc(file_path, index=index, use_ln=True, radius=100)


    # for cr in [1, 10, 20, 40, 60, 80, 100]:
    #     tc, _ = ap.thermo.basic.get_tc(da2=0.077 / (0.0140 ** 2) * 3600 * 24 * 365.24, sda2=0, E=47*1000 * 4.184, sE=0, pho=0,
    #                                    cooling_rate=cr, A=55)
    #     print(f"cooling rate = {cr}, tc = {tc}")
    #
    # for cr in [1, 10, 20, 40, 60, 80, 100]:
    #     tc, _ = ap.thermo.basic.get_tc(da2=18 * 3600 * 24 * 365.24, sda2=0, E=87*1000 * 4.184, sE=0, pho=0,
    #                                    cooling_rate=cr, A=6)
    #     print(f"cooling rate = {cr}, tc = {tc}")
    #
    # for cr in [1, 10, 20, 40, 60, 80, 100]:
    #     tc, _ = ap.thermo.basic.get_tc(da2=10**9.05 * 3600 * 24 * 365.24, sda2=0, E=83.79*1000 * 4.184, sE=0, pho=0,
    #                                    cooling_rate=cr, A=55)
    #     print(f"cooling rate = {cr}, tc = {tc}")

    export_ar_relase_curve()




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - err.py
# ==========================================
#
# This package contains error propagation functions
#
"""

import numpy as np
import pandas as pd


def add(*args: float) -> float:
    """
        For Y = X1 +/- X2 +/- ... +/- Xn
    Args:
        *args:errors in 1 sigma

    Returns: float, propagated error

    """
    k = sum([i ** 2 for i in args]) ** .5
    return k


def mul(*args: tuple) -> float:
    """
        For Y = X1 * X2 * ... * Xn
    Args:
        *args: tuple, (v1, s1), (v2, s2), ...

    Returns: float, propagated error

    """
    # k = abs(np.prod([arg[0] for arg in args])) * sum([np.divide(arg[1] ** 2, arg[0] ** 2) for arg in args]) ** .5
    v = np.ma.array(args, mask=False)
    k = 0
    for i in range(v.shape[0]):
        v.mask[i, 0] = True
        k += np.prod(v[:, 0]) ** 2 * v[i, 1] ** 2
        v.mask[i, 0] = False

    return np.sqrt(k)


def rec(a: tuple) -> float:
    """
        For Y = 1 / X
    Args:
        *a: tuple, (value, error)

    Returns: float, propagated error

    """
    if isinstance(a, pd.Series):
        a = a.values.tolist()
    k = np.divide(abs(a[1]), a[0] ** 2)
    return k


def div(*args: tuple) -> float:
    """
        For Y = X1 / X2 / ... / Xn
    Args:
        *args: float, (v1, s1), (v2, s2), ...

    Returns: float, propagated error

    """
    args = [arg if index == 0 else (np.divide(1, arg[0]), rec(arg)) for index, arg in enumerate(args)]
    k = mul(*args)
    return k


def pow(a0: tuple, a1: tuple):
    """
        For y = pow(a0, a1), y = a0 ^ a1
    Args:
        a0: tuple, (value, error)
        a1: tuple, (value, error)

    Returns: float, propagated error

    """
    p1 = a0[1] ** 2 * (a1[0] * a0[0] ** (a1[0] - 1)) ** 2
    p2 = a1[1] ** 2 * (a0[0] ** a1[0] * np.log(a0[0])) ** 2
    k = (p1 + p2) ** .5
    return k


def log(a: tuple) -> float:
    """
        For y = ln(a)
    Args:
        a: tuple, (value, error)

    Returns: float, propagated error

    """
    k = np.divide(abs(a[1]), abs(a[0]))
    return k


def cor(sX: float, sY: float, sZ: float):
    """
    Calculate correlation coefficient of errors.

    Parameters
    ----------
    sX : relative error of X, where X/Z vs. Y/Z
    sY : relative error of Y
    sZ : relative error of Z

    Returns
    -------
    """
    if sZ == 0:
        return np.nan
    k = np.divide(sZ ** 2, ((sZ ** 2 + sX ** 2) * (sZ ** 2 + sY ** 2)) ** .5)
    return k



#  Copyright (C) 2024 Yang. - All Rights Reserved

# !/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2024 Yang 
# ararpy - info
# ==========================================
#
#
# 
"""


def name(smp, n: str = None):
    if n is None:
        return smp.Info.sample.name
    elif isinstance(n, str):
        smp.Info.sample.name = n
        return n
    else:
        raise ValueError(f"{n} is not a string")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# webarar - raw_file
# ==========================================
#
#
#
"""

from typing import List, Union
import traceback
import os
import re
import pickle
import chardet
from xlrd import open_workbook
from datetime import datetime
from parse import parse as string_parser
import dateutil.parser as datetime_parser
from ..calc.arr import get_item
from ..calc.basic import utc_dt
from ..smp.sample import RawData

""" Open raw data file """

DEFAULT_SAMPLE_INFO = {
    "ExpName": "en",        # "Experiment Name"
    "StepName": "sn",       # "Step Name"
    "StepType": "st",       # "Step Type"
    "StepLabel": "sl",      # "Step Label"
    "ZeroYear": "date",     # "Zero Date Year"
    "ZeroHour": "time",     # "Zero Time Hour"
    "ZeroMon": "mon",       # "Zero Date Month"
    "ZeroMin": "min",       # "Zero Time Minute"
    "ZeroDay": "day",       # "Zero Date Day"
    "ZeroSec": "sec",       # "Zero Time Second"
    "SmpName": "smpn",      # "Sample Name"
    "SmpLoc": "smpl",       # "Sample location"
    "SmpMatr": "smpm",      # "Sample Material"
    "ExpType": "et",        # "Experiment Type"
    "SmpWeight": "smpw",    # "Sample Weight"
    "Stepunit": "su",       # "Step unit"
    "HeatingTime": "ht",    # "Heating time"
    "InstrName": "inn",     # "Instrument name"
    "Researcher": "re",     # "Researcher"
    "Analyst": "an",        # "Analyst"
    "Lab": "la",            # "Laboratory"
    "Jv": "jv",             # "J value"
    "Jsig": "je",           # "J value error"
    "MDF": "mdf",           # "MDF"
    "MDFSig": "mdfe",       # "MDF error"
    "CalcName": "cp",       # "Calc params"
    "IrraName": "in",       # "Irra name"
    "IrraLabel": "il",      # "Irra label"
    "IrraPosH": "ih",       # "Irra position H"
    "IrraPosX": "ix",       # "Irra position X"
    "IrraPosY": "iy",       # "Irra position Y"
    "StdName": "stdn",      # "Standard name"
    "StdAge": "stda",       # "Standard age"
    "StdAgeSig": "stde",    # "Standard age error"
}



def open_file(file_path: str, input_filter: List[Union[str, int, bool]], name=None):
    """
    Parameters
    ----------
    file_path:
    input_filter

    Returns
    -------
    step_list -> [[[header of step one], [cycle one in the step], [cycle two in the step]],[[],[]]]
        example:
            [
                [
                    [1, '6/8/2019  8:20:51 PM', 'BLK', 'B'],  # step header
                    # [sequence index, date time, label, value]
                    ['1', 12.30, 87.73, 12.30, 1.30, 12.30, 0.40, 12.30, 0.40, 12.30, 0.44],  # step/sequence 1
                    # [cycle index, time, Ar40 signal, time, Ar39 signal, ..., time, Ar36 signal]
                    ['2', 24.66, 87.70, 24.66, 1.14, 24.66, 0.36, 24.66, 0.35, 24.66, 0.43],  # step/sequence 2
                    ...
                    ['10', 123.06, 22262.68, 123.06, 6.54, 123.06, 8.29, 123.06, 0.28, 123.06, 29.22],
                ],
                [
                    ...
                ]
            ]

    """
    extension = str(os.path.split(file_path)[-1]).split('.')[-1]
    try:
        handler = {'txt': open_raw_txt, 'excel': open_raw_xls, 'Seq': open_raw_seq}[
            ['txt', 'excel', 'Qtegra Excel (deleted)', 'Seq'][int(input_filter[1])]]
    except KeyError:
        raise FileNotFoundError("Wrong File.")
    return handler(file_path, input_filter, name)


def open_raw_txt(file_path, input_filter: List[Union[str, int]], name=None):
    """
    Parameters
    ----------
    input_filter
    file_path

    Returns
    -------

    """
    if not input_filter:
        raise ValueError("Input filter is empty array.")

    if os.path.splitext(file_path)[1][1:].lower() != input_filter[0].strip().lower():
        raise ValueError("The file does not comply with the extension in the given filter.")

    with open(file_path, 'rb') as f:
        contents = f.read()
        encoding = chardet.detect(contents)
        lines = [line.strip().split(['\t', ';', " ", ",", input_filter[3]][int(input_filter[2])])
                 for line in contents.decode(encoding=encoding["encoding"]).split('\r\n')]

    file_name = os.path.basename(file_path).rstrip(os.path.splitext(file_path)[-1]) if name is None else name
    step_list = get_raw_data([lines], input_filter, file_name=file_name)
    return {'data': step_list}


def open_raw_xls(file_path, input_filter: List[Union[str, int]], name):
    """
    Parameters
    ----------
    file_path
    input_filter

    Returns
    -------

    """
    if not input_filter:
        raise ValueError("Input filter is empty array.")

    if os.path.splitext(file_path)[1][1:].lower() != input_filter[0].strip().lower():
        raise ValueError("The file does not comply with the extension in the given filter.")

    def _get_content_from_sheet(_index) -> List[List[Union[str, bool, int, float]]]:
        try:
            _sheet = wb.sheet_by_index(_index)
        except IndexError:
            return []
        else:
            return [[_sheet.cell(_row, _col).value for _col in range(_sheet.ncols)] for _row in range(_sheet.nrows)]

    wb = open_workbook(file_path)
    contents = [_get_content_from_sheet(i) for i in range(100)]
    file_name = os.path.basename(file_path).rstrip(os.path.splitext(file_path)[-1]) if name is None else name
    step_list = get_raw_data(contents, input_filter, file_name=file_name)

    return {'data': step_list}


def open_raw_seq(file_path, input_filter=None, name=None):

    class RenameUnpickler(pickle.Unpickler):
        def find_class(self, module: str, name: str):
            MODULE = RawData().__module__
            renamed_module = module
            if '.sample' in module and module != MODULE:
                renamed_module = MODULE
            try:
                return super(RenameUnpickler, self).find_class(renamed_module, name)
            except AttributeError:
                return super(RenameUnpickler, self).find_class(renamed_module, 'ArArBasic')

    def renamed_load(file_obj):
        return RenameUnpickler(file_obj).load()

    with open(file_path, 'rb') as f:
        sequences = renamed_load(f)

    # with open(file_path, 'rb') as f:
    #     sequences = pickle.load(f)
    #     print(sequences[0].__module__)

    name_list = []
    for seq in sequences:
        while seq.name in name_list:
            seq.name = f"{seq.name}-{seq.index}"
        name_list.append(seq.name)

    return {'sequences': sequences}


def get_raw_data(file_contents: List[List[Union[int, float, str, bool, list]]], input_filter: list,
                 file_name: str = "") -> list:
    """
    Parameters
    ----------
    file_name
    file_contents
    input_filter

    Returns
    -------

    """

    def datetime_parse(string: str, f: str):
        try:
            return datetime.strptime(string, f)
        except ValueError as v:
            print(traceback.format_exc())
            if f.strip() == "":
                return datetime_parser.parse(string)
            elif len(v.args) > 0 and v.args[0].startswith('unconverted data remains: '):
                # %f handles microseconds (6 digits), remove the remains
                return datetime_parse(string[:-(len(v.args[0]) - 26)], f)
            # elif f.upper()[-2:] in ['PM', 'AM']:
            #     return datetime_parse(string, f[:-2])
            else:
                raise

    def step_num_parse(s: str):
        try:
            return f"{round(float(s)):02d}"
        except (ValueError, TypeError):
            return s

    step_list = []
    idx = 0  # step_index

    header = input_filter[5]
    isotope_index = input_filter[8:28]
    data_index = input_filter[4:33]
    strings_index = input_filter[33:41]         # strings: filename, parsing information, datetime, timezone
    optional_info_index = input_filter[41:-7]       # from Exp Name to Std Age Error
    check_box_index = input_filter[-7:]

    while True:  # measurment steps sloop

        # ============ all text information ============
        options = get_sample_info(file_contents, optional_info_index, default="", base=[1, 1 - idx, 1])

        # ============ Step name ============
        try:
            step_name = options.get('StepName')
            experiment_name = options.get('ExpName')
            if check_box_index[1] and strings_index[0].strip() != "":
                res = string_parser(strings_index[0], file_name)
                if res is not None and "en" in res.named.keys():
                    experiment_name = res.named.get("en")
                if res is not None and "sn" in res.named.keys():
                    step_name = res.named.get("sn")
            if str(step_name) == "":
                raise ValueError(f"Step name not found, break reading steps")
            step_name = f"{experiment_name}-{step_num_parse(step_name)}"
        except (TypeError, ValueError, IndexError):
            # When parsing the step name fails, the end of the file has been reached
            # raise
            print(traceback.format_exc())
            break
        else:
            options.update({'StepName': step_name})
            options.update({'ExpName': experiment_name})

        # ============ Step information ============
        try:
            if check_box_index[2]:
                string = get_item(file_contents, strings_index[1:4], default="", base=[1, 1 - idx, 1])
                res = string_parser(strings_index[4], string)
                if res is not None:
                    options.update(dict(zip(DEFAULT_SAMPLE_INFO.keys(), [res.named.get(value, options.get(key)) for key, value in DEFAULT_SAMPLE_INFO.items()])))
        except (TypeError, ValueError, IndexError):
            print(traceback.format_exc())
            break
        else:
            pass

        # ============ Zero datetime ============
        timezone = strings_index[7] if strings_index[7] != "" else "utc"
        options.update({'TimeZone': timezone})
        try:
            if check_box_index[3]:  # date in one string
                zero_date = datetime_parse(options.get('ZeroYear'), strings_index[5])
            else:
                zero_date = datetime(year=options.get('ZeroYear'), month=options.get('ZeroMon'),
                                     day=options.get('ZeroDay'))

            if check_box_index[4]:  # time in one string
                zero_time = datetime_parse(options.get('ZeroHour'), strings_index[6])
            else:
                zero_time = datetime(year=2020, month=12, day=31, hour=options.get('ZeroHour'),
                                     minute=options.get('ZeroMin'), second=options.get('ZeroSec'))

            zero_datetime = datetime(
                zero_date.year, zero_date.month, zero_date.day, zero_time.hour, zero_time.minute, zero_time.second)
            # adjust to UTC
            zero_datetime = utc_dt(zero_datetime, tz=timezone).isoformat(timespec='seconds')
            options.update({'ZeroDT': zero_datetime})
        except (TypeError, ValueError, IndexError) as e:
            print(traceback.format_exc())
            raise ValueError(f"Failed to parse zero datetime: {e}")
            # zero_datetime = datetime(1970, 1, 1, 0, 0, 0).isoformat(timespec='seconds')
        current_step = [[step_name, zero_datetime, options]]

        # ============ isotope data ============
        break_num = 0
        cycle_num = 0
        f = float(data_index[27])  # Intensity Scale Factor
        data_content = file_contents[data_index[0] - 1 if data_index[0] != 0 else 0]
        base = 1
        for i in range(2000):  # measurement cycle sloop
            if break_num < data_index[25]:
                break_num += 1
                continue
            break_num = 0
            if int(data_index[2]) == 0:  # == 0, vertical
                start_row = data_index[24] * cycle_num + data_index[25] * cycle_num + header + idx
                try:
                    current_step.append([
                        str(cycle_num + 1),
                        # in sequence: Ar36, Ar37, Ar38, Ar39, Ar40
                        float(data_content[start_row + isotope_index[18] - base][isotope_index[19] - base]),
                        float(data_content[start_row + isotope_index[16] - base][isotope_index[17] - base]) * f,
                        float(data_content[start_row + isotope_index[14] - base][isotope_index[15] - base]),
                        float(data_content[start_row + isotope_index[12] - base][isotope_index[13] - base]) * f,
                        float(data_content[start_row + isotope_index[10] - base][isotope_index[11] - base]),
                        float(data_content[start_row + isotope_index[ 8] - base][isotope_index[ 9] - base]) * f,
                        float(data_content[start_row + isotope_index[ 6] - base][isotope_index[ 7] - base]),
                        float(data_content[start_row + isotope_index[ 4] - base][isotope_index[ 5] - base]) * f,
                        float(data_content[start_row + isotope_index[ 2] - base][isotope_index[ 3] - base]),
                        float(data_content[start_row + isotope_index[ 0] - base][isotope_index[ 1] - base]) * f,
                    ])
                except (ValueError, IndexError):
                    print(f"Cannot parse isotope data")
                    print(traceback.format_exc())
                    current_step.append([
                        str(cycle_num + 1), None, None, None, None, None, None, None, None, None, None,
                    ])
            elif int(data_index[2]) == 1:  # == 1, horizontal
                start_row = data_index[1] + idx
                col_inc = data_index[24] * cycle_num + data_index[25] * cycle_num - base
                try:
                    current_step.append([
                        str(cycle_num + 1),
                        # Ar36, Ar37, Ar38, Ar39, Ar40
                        float(data_content[start_row][isotope_index[19] + col_inc]),
                        float(data_content[start_row][isotope_index[17] + col_inc]) * f,
                        float(data_content[start_row][isotope_index[15] + col_inc]),
                        float(data_content[start_row][isotope_index[13] + col_inc]) * f,
                        float(data_content[start_row][isotope_index[11] + col_inc]),
                        float(data_content[start_row][isotope_index[ 9] + col_inc]) * f,
                        float(data_content[start_row][isotope_index[ 7] + col_inc]),
                        float(data_content[start_row][isotope_index[ 5] + col_inc]) * f,
                        float(data_content[start_row][isotope_index[ 3] + col_inc]),
                        float(data_content[start_row][isotope_index[ 1] + col_inc]) * f,
                    ])
                except (ValueError, IndexError):
                    print(f"Cannot parse isotope data")
                    print(traceback.format_exc())
                    current_step.append([
                        str(cycle_num + 1), None, None, None, None, None, None, None, None, None, None,
                    ])
            else:
                raise ValueError(f"{data_index[2]} not in [0, 1]")

            cycle_num += 1
            if cycle_num >= data_index[3]:
                break

        step_list.append(current_step)
        idx = data_index[28] * len(step_list)
        if not check_box_index[0] or len(step_list) >= 500:  # check_box_index[0]: multiple sequences
            # print(f"Multiple Sequence = {check_box_index[0]}, Step number = {len(step_list)}")
            break

    if not step_list:
        raise ValueError("Failed to read the original file. It might be because the names of the experiments or steps were not recognized.")

    return step_list


def get_sample_info(file_contents: list, index_list: list, default="", base: Union[int, tuple, list] = 1) -> dict:
    """
    Parameters
    ----------
    file_contents
    index_list
    default
    base

    Returns
    -------

    """
    sample_info = DEFAULT_SAMPLE_INFO.copy()
    sample_info.update({
        "ExpName": get_item(file_contents, index_list[0:3], default=default, base=base),
        "StepName": get_item(file_contents, index_list[3:6], default=default, base=base),
        "StepType": get_item(file_contents, index_list[6:9], default=default, base=base),
        "StepLabel": get_item(file_contents, index_list[9:12], default=default, base=base),
        "ZeroYear": get_item(file_contents, index_list[12:15], default=default, base=base),
        "ZeroHour": get_item(file_contents, index_list[15:18], default=default, base=base),
        "ZeroMon": get_item(file_contents, index_list[18:21], default=default, base=base),
        "ZeroMin": get_item(file_contents, index_list[21:24], default=default, base=base),
        "ZeroDay": get_item(file_contents, index_list[24:27], default=default, base=base),
        "ZeroSec": get_item(file_contents, index_list[27:30], default=default, base=base),
        "SmpName": get_item(file_contents, index_list[30:33], default=default, base=base),
        "SmpLoc": get_item(file_contents, index_list[33:36], default=default, base=base),
        "SmpMatr": get_item(file_contents, index_list[36:39], default=default, base=base),
        "ExpType": get_item(file_contents, index_list[39:42], default=default, base=base),
        "SmpWeight": get_item(file_contents, index_list[42:45], default=default, base=base),
        "Stepunit": get_item(file_contents, index_list[45:48], default=default, base=base),
        "HeatingTime": get_item(file_contents, index_list[48:51], default=default, base=base),
        "InstrName": get_item(file_contents, index_list[51:54], default=default, base=base),
        "Researcher": get_item(file_contents, index_list[54:57], default=default, base=base),
        "Analyst": get_item(file_contents, index_list[57:60], default=default, base=base),
        "Lab": get_item(file_contents, index_list[60:63], default=default, base=base),
        "Jv": get_item(file_contents, index_list[63:66], default=default, base=base),
        "Jsig": get_item(file_contents, index_list[66:69], default=default, base=base),
        "MDF": get_item(file_contents, index_list[69:72], default=default, base=base),
        "MDFSig": get_item(file_contents, index_list[72:75], default=default, base=base),
        "CalcName": get_item(file_contents, index_list[75:78], default=default, base=base),
        "IrraName": get_item(file_contents, index_list[78:81], default=default, base=base),
        "IrraLabel": get_item(file_contents, index_list[81:84], default=default, base=base),
        "IrraPosH": get_item(file_contents, index_list[84:87], default=default, base=base),
        "IrraPosX": get_item(file_contents, index_list[87:90], default=default, base=base),
        "IrraPosY": get_item(file_contents, index_list[90:93], default=default, base=base),
        "StdName": get_item(file_contents, index_list[93:96], default=default, base=base),
        "StdAge": get_item(file_contents, index_list[96:99], default=default, base=base),
        "StdAgeSig": get_item(file_contents, index_list[99:102], default=default, base=base),
    })
    return sample_info



#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Check arr
# ==========================================
#
#
# 
"""
import os
import ararpy as ap
import numpy as np
import pdf_maker as pm


def read_sample(arr_file_path):
    # 读取样品信息
    if not arr_file_path.endswith('.arr'):
        for root, dirs, files in os.walk(arr_file_path):
            for file in files:
                if file.endswith('.arr'):
                    arr_file_path = os.path.join(arr_file_path, file)
                    break
    print(f"arr file: {arr_file_path}")
    sample = ap.from_arr(file_path=arr_file_path)
    name = sample.name()
    sequence = sample.sequence()
    nsteps = sequence.size
    arratio = sample.TotalParam[0][0]
    if sample.Info.sample.type == "Unknown":
        print(f"{name = }, {arratio = }")
        # sample.TotalParam[0] = [298.56 for i in sample.TotalParam[0]]
        sample.Info.preference["confidenceLevel"] = 2
        # ap.smp.calculation.recalculate(sample, re_plot=True, re_plot_style=True, re_set_table=True, re_table_style=True)
        ap.save(sample, arr_file_path)


loc = r"C:\Users\Young\OneDrive\00-Projects\【2】个人项目\2022-05论文课题\【3】分析测试\ArAr\01-VU实验数据和记录\Arr Data"
# read_sample(os.path.join(loc, "20240918_24FY87.arr"))
# read_sample(loc)

for root, dirs, files in os.walk(loc):
    for file in files:
        if file.endswith('.arr'):
            arr_file_path = os.path.join(loc, file)
            read_sample(arr_file_path)





#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Granite Cooling History
# ==========================================
#
#
# 
"""

#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Show MDD results
# ==========================================
#
#
# 
"""
import ararpy as ap
import numpy as np
import pdf_maker as pm
import os

import matplotlib
from matplotlib.collections import PathCollection

matplotlib.use('TkAgg')
matplotlib.rc('font',family='Arial', size=10)
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
# 设置全局字体，确保中文正常显示
# plt.rcParams["font.family"] = ["SimHei"]  # 中文字体
# plt.rcParams['axes.unicode_minus'] = False  # 解决负号显示问题



def conf(input_x, input_y, num=None, using_binom=False, using_normal=False, start=None, end=None):
    """
    Calculate 90% confident interval of the given distribution of cooling histories.
    Parameters
    ----------
    input_x: x, array like 2D
    input_y: y, array like 2D
    count:

    Returns
    -------

    """

    if len(input_x) != len(input_y):
        raise ValueError("length of input x does not equal to that of input y")

    ns = len(input_x)
    if start is None:
        start = min([min(i) for i in input_x])
    if end is None:
        end = max([max(i) for i in input_x])
    nt = 18 if num is None else num

    if not (using_normal or using_binom):
        using_binom = True

    x_conf = np.linspace(start, end, num=nt)
    y_conf = np.zeros((nt, 4))
    ytemp = np.zeros((nt, ns))
    # 初始化数组
    xmed = np.zeros(nt)
    ave = np.zeros(nt)
    adev = np.zeros(nt)
    sdev = np.zeros(nt)
    var = np.zeros(nt)
    skew = np.zeros(nt)
    curt = np.zeros(nt)

    # 计算每个曲线的温度值
    for k, out_x in enumerate(x_conf):
        for i in range(ns):
            xs = input_x[i]
            ys = input_y[i]
            for j in range(len(xs)-1):
                if xs[j] >= out_x >= xs[j+1]:
                    ytemp[k, i] = (ys[j+1] - ys[j]) / (xs[j+1] - xs[j]) * (out_x - xs[j]) + ys[j]
                    break
        # 排序
        ytemp[k] = np.sort(ytemp[k])
        # 计算均值、中位数
        ave[k], adev[k], sdev[k], var[k], skew[k], curt[k], xmed[k] = stats = moment(ytemp[k])
        # 计算置信区间索引
        if using_binom:
            j1, j2 = ap.smp.diffusion_funcs.binom(ns)  #
            j3, j4 = int(round(ns * 0.05)) - 1, int(ns - round(ns * 0.05)) - 1  # 95%
            y1, y2, y3, y4 = ytemp[k][j1], ytemp[k][j2], ytemp[k][j3], ytemp[k][j4]
        elif using_normal:
            y1, y2, y3, y4 = ave[k] - sdev[k], ave[k] + sdev[k], ave[k] - 2 * sdev[k], ave[k] + 2 * sdev[k]
        else:
            raise KeyError("use binom or using normal")
        y_conf[k] = [y1, y2, y3, y4]
    return x_conf, y_conf


def moment(data):
    """ 计算分布的统计矩阵：均值、平均偏差、标准差、方差、偏度、峰度、中位数 """
    # 去除NaN值
    data = data[~np.isnan(data)]
    n = len(data)
    if n == 0:
        return [np.nan] * 7

    mean = np.mean(data)
    adev = np.mean(np.abs(data - mean))  # 平均偏差
    sdev = np.std(data, ddof=1)  # 样本标准差
    var = sdev ** 2  # 方差
    med = np.median(data)  # 中位数

    # 偏度 (三阶矩)
    if sdev == 0:
        skew = 0.0
    else:
        skew = np.sum(((data - mean) / sdev) ** 3) / n

    # 峰度 (四阶矩)
    if sdev == 0:
        curt = 0.0
    else:
        curt = np.sum(((data - mean) / sdev) ** 4) / n - 3.0  # 减去3使正态分布峰度为0

    return mean, adev, sdev, var, skew, curt, med


def is_number(s):
    try:
        float(s)
        return True
    except:
        return False


def read_data(file):
    data = []
    rows = []
    go = True
    while go:
        go = False
        group = np.zeros([200, 20], dtype='<U64')
        row = 0
        while True:
            try:
                line = file.readline().rstrip("\n")
                print(line)
                if line == "":
                    break
                for col, _ in enumerate([str(l) for l in line.split(',')]):
                    group[row][col] = _
            except (Exception, BaseException) as e:
                print(e)
                break
            else:
                row += 1
        data.append(group)
        rows.append(row)
    return data, rows


def plot_CH(file, sname=None, title="", **params):

    file_IN = open(file, "r")
    [data], [nstep] = read_data(file_IN)


    initial_age = params.pop("initial_age", 0)
    final_age = params.pop("final_age", 20)
    x1_extent = params.pop("x1_extent", [0, 50])
    y1_extent = params.pop("y1_extent", [0, 900])
    x2_extent = params.pop("x2_extent", [initial_age, final_age])
    y2_extent = params.pop("y2_extent", [0, 500])
    x3_extent = params.pop("x3_extent", [4, 16])
    y3_extent = params.pop("y3_extent", [-10, 0])
    x4_extent = params.pop("x4_extent", [0, 1000])
    y4_extent = params.pop("y4_extent", [200, 1600])
    autoscaling = params.pop("autoscaling", False)
    export_PDF = params.pop("export_PDF", False)
    index = params.pop("index", [])

    main_color = ['#397DA1', '#BA5624', '#212121', '#3C6933', '#74757C', '#214193', '#524715']  # blue red black green
    middle_color = ['#83CDFA', '#F1B595', '#737373', '#84B775', '#C1C3CB', '#6D8AE1', '#C0A737']
    shallow_color = ['#E0F1FE', '#FBEBE3', '#DDDDDD', '#CFFEC0', '#EDEEF8', '#D6DFFC', '#FCF0BC']

    print(f"{sname = }, {title = }")

    fig, axs = plt.subplots(1, 1, figsize=(6, 4))

    font_settings = {
        # 'fontsize': 10,
        # 'fontfamily': 'Arial'
    }

    colors = {
        "biotite": {"facecolor": "#F1B595", "edgecolor": "#BA5624"},
        "黑云母": {"facecolor": "#F1B595", "edgecolor": "#BA5624"},
        "zircon": {"facecolor": "#737373", "edgecolor": "#212121"},
        "锆石": {"facecolor": "#737373", "edgecolor": "#212121"},
        "muscovite": {"facecolor": "#83CDFA", "edgecolor": "#397DA1"},
        "白云母": {"facecolor": "#83CDFA", "edgecolor": "#397DA1"},
        "monazite": {"facecolor": "#84B775", "edgecolor": "#3C6933"},
        "独居石": {"facecolor": "#84B775", "edgecolor": "#3C6933"},
        "apatite": {"facecolor": "#D6DFFC", "edgecolor": "#6D8AE1"},
        "磷灰石": {"facecolor": "#D6DFFC", "edgecolor": "#6D8AE1"},
        "xenotime": {"facecolor": "#C0A737", "edgecolor": "#524715"},
        "磷钇矿": {"facecolor": "#C0A737", "edgecolor": "#524715"},
        "电气石": {"facecolor": "white", "edgecolor": "#524715"},
        "角闪石": {"facecolor": "white", "edgecolor": "#524715"},
    }

    axs.set_xlim(*x1_extent, auto=autoscaling)
    axs.set_ylim(*y1_extent, auto=autoscaling)
    body = data[:nstep, 0]
    area = data[:nstep, 1]
    lithology = data[:nstep, 2]
    sample_number = data[:nstep, 3]
    mineral = data[:nstep, 4]
    method = data[:nstep, 5]
    age = data[:nstep, 6].astype(np.float32)
    sage = data[:nstep, 7].astype(np.float32)
    reference = data[:nstep, 8]
    tc = data[:nstep, 9].astype(np.float32)
    stc = data[:nstep, 10].astype(np.float32)
    print(f"{nstep = }")
    for i in np.argsort(sage)[::-1]:
        color = colors[mineral[i]]
        print(f"age = {age[i]} +/- {sage[i]}, Tc = {tc[i]} +/- {stc[i]}")
        linestyle = "solid"
        rect = Rectangle(
            (age[i]-sage[i], tc[i]-stc[i]), width=2 * sage[i], height=2 * stc[i], **color,
            linestyle=linestyle
        )
        axs.add_patch(rect)
    axs.set_title(f'{title}', loc='center', y=1, **font_settings)
    axs.set_xlabel(f'Age (Ma)', **font_settings)
    axs.set_ylabel(f'Temperature (°C)', **font_settings)

    fig.tight_layout()
    plt.show()

    if not export_PDF:
        return

    filename = f"{title} - cooling history - matplotlib"

    params_list = {
        "page_size": 'a4', "ppi": 72, "width": 14, "height": 8,
        "pt_width": 0.8, "pt_height": 0.8, "pt_left": 0.16, "pt_bottom": 0.18,
        "offset_top": 0, "offset_right": 0, "offset_bottom": 20, "offset_left": 30,
        "plot_together": False, "show_frame": False,
        'xlabel_offset': 8, 'ylabel_offset': 2
    }

    plot_data = {
        "data": [
            transform(axs),
        ],
        "file_name": filename,
        "plot_names": [f"plotname"],
    }

    filepath = os.path.join(r"C:\Users\Young\Downloads", f"{filename}.pdf")
    cvs = [[ap.smp.export.get_cv_from_dict(plot, **params_list) for plot in plot_data['data']]]
    for i in range(len(cvs[0])):
        pt = cvs[0][i]._plot_areas[0]
        title = pt.get_comp(comp_name="title")
        title._y -= 2
        title._z_index = 999
    filepath = ap.smp.export.export_chart_to_pdf(cvs, filename, filepath)


def transform(ax: plt.Axes):
    xlabels = [i.get_text().replace('−', '-') for i in ax.get_xticklabels()]
    ylabels = [i.get_text().replace('−', '-') for i in ax.get_yticklabels()]
    linestyles = {'-': 'solid', '--': 'dashed', '-.': 'dashdot', ':': 'dotted'}

    series = []
    for i, line in enumerate(ax.lines):
        xy_data = line.get_xydata()  # [[x1, y1], [x2, y2], ...]
        line_style = linestyles.get(line.get_linestyle(), 'solid')
        series.append({
            'type': 'series.line', 'id': f'line-{i}', 'name': f'line-{i}',
            'color': line.get_color(), 'line_width': 1, 'line_style': line_style,
            'data': xy_data, 'line_caps': 'none'
        })
        if bool(line._marker):
            series.append({
                'type': 'series.scatter', 'id': f'line-marker-{i}', 'name': f'line-marker-{i}',
                'stroke_color': line.get_markeredgecolor(), 'fill_color': line.get_markerfacecolor(),
                'data': xy_data, 'size': 2,
                # 'symbol': line._marker.markers.get(line.get_marker(), 'square'),
                'symbol': 'rec'
            })
    for i, collection in enumerate(ax.collections):
        series.append({
            'type': 'series.scatter', 'id': f'scatter-{i}', 'name': f'{collection.get_label()}',
            'stroke_color': collection.get_edgecolor()[0][:3], 'fill_color': collection.get_edgecolor()[0][:3],
            'data': collection.get_offsets(), 'size': 2,
            'symbol': 'rec'
        })

    for i, text in enumerate(ax.texts):
        xy_data = text.get_position()  # [[x1, y1], [x2, y2], ...]
        series.append({
            'type': 'series.text', 'id': f'text-{i}', 'name': f'text-{i}',
            'color': text.get_color(), 'data': [xy_data], 'text': text.get_text().replace('\n', '<r>'),
            'size': 8
        })

    for child in ax._children:
        if isinstance(child, Rectangle):
            ld = [*child.get_xy(), child.get_width(), child.get_height()]
            series.append({
                'type': 'series.rect', 'id': f'rect-{child.get_gid()}', 'name': f'rect-{child.get_gid()}',
                'color': child.get_edgecolor()[:3], 'fill_color': child.get_facecolor()[:3], 'fill': True,
                'data': [ld], 'label': child.get_label(), 'clip': True
            })

    series.append({
        'type': 'series.text', 'id': f'title', 'name': f'title',
        'color': 'black', 'data': [[sum(ax.get_xlim()) / 2, ax.get_ylim()[1]]],
        'h_align': "middle", 'v_align': "top",
        'text': ax.get_title(), 'size': 8
    })

    if ax.legend_ is not None:
        for handle, text in zip(ax.legend_.legend_handles, ax.legend_.texts):
            series.append({
                'type': 'series.text', 'id': f'legend', 'name': f'legend',
                'color': text.get_color(), 'data': [[ax.get_xlim()[0], ax.get_ylim()[0]]],
                'h_align': "left", 'v_align': "bottom",
                'text': text.get_text(), 'size': 8
            })
            if isinstance(handle, plt.Line2D):
                series.append({
                    'type': 'series.line', 'id': f'legend-line', 'name': f'legend-line-{text.get_text()}',
                    'color': handle.get_color(), 'data':[[ax.get_xlim()[0], ax.get_ylim()[0]], [ax.get_xlim()[1], ax.get_ylim()[1]]],
                    'line_width': 1, 'line_style': linestyles.get(handle.get_linestyle(), 'solid')
                })
            if isinstance(handle, PathCollection):
                stroke_c = handle.get_edgecolor()[0][:3]
                stroke_c = f"#{int(stroke_c[0]*255):02x}{int(stroke_c[1]*255):02x}{int(stroke_c[2]*255):02x}"
                fill_c = handle.get_facecolor()[0][:3]
                fill_c = f"#{int(fill_c[0]*255):02x}{int(fill_c[1]*255):02x}{int(fill_c[2]*255):02x}"
                series.append({
                    'type': 'series.scatter', 'id': f'legend-scatter', 'name': f'legend-scatter-{text.get_text()}',
                    'stroke_color': stroke_c, 'fill_color': fill_c,
                    'data': [[sum(ax.get_xlim()) / 2, sum(ax.get_ylim()) / 2]],
                    'size': 2, 'symbol': 'rec'
                })

    data = {
        'xAxis': [{
            'extent': ax.get_xlim(), 'interval': xlabels, 'title': ax.get_xlabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'yAxis': [{
            'extent': ax.get_ylim(), 'interval': ylabels, 'title': ax.get_ylabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'series': series
    }

    # print(data)
    return data



if __name__ == "__main__":

    params = {
        "initial_age": 0,
        "final_age": 11,
        "x1_extent": [0, 60],
        "y1_extent": [0, 900],
        "x2_extent": [0, 11],
        "y2_extent": [0, 500],
        "x3_extent": [4, 16],
        "y3_extent": [-10, 0],
        "x4_extent": [0, 1800],
        "y4_extent": [200, 1600],
        "autoscaling": False,
        # "autoscaling": True,
        # "export_PDF": False,
        "export_PDF": True,
        "index": list(range(2, 17)),
    }


    s = 'yalaxiangbo'
    s = 'paiku'
    s = 'kuday'
    s = 'ramba'
    s = 'gaowu'
    s = 'luozha'
    loc = os.path.join("D:\DjangoProjects\webarar\private\mdd", f"{s}.txt")
    plot_CH(file=loc, title=f"{s.capitalize()} - Cooling History", **params)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - basic
# ==========================================
#
#
#
"""
import copy
import secrets
import numpy as np
from scipy import stats
from datetime import datetime, timezone, timedelta
import pytz


def utc_dt(dt: datetime, tz: str = "utc", is_dst: bool = False) -> datetime:
    """
    Parameters
    ----------
    dt
    tz
    is_dst: only valid when an ambiguous time inputted

    Returns
    -------

    """
    tz = pytz.timezone(tz)
    return tz.localize(dt, is_dst).astimezone(pytz.utc)


def get_datetime(t_year: int, t_month: int, t_day: int, t_hour: int, t_min: int, t_seconds: int = 0,
                 tz_hour: int = 0, tz_min: int = 0, base=None):
    """
    :param t_year: int
    :param t_month: int
    :param t_day: int
    :param t_hour: int
    :param t_min: int
    :param t_seconds: int, default == 0
    :param tz_hour: int, default == 0
    :param tz_min: int, default == 0
    :param base: base time [y, m, d, h, m]
    :return: seconds since 1970-1-1 0:00
    """
    t_year, t_month, t_day, t_hour, t_min, t_seconds, tz_hour, tz_min = \
        int(t_year), int(t_month), int(t_day), int(t_hour), int(t_min), int(t_seconds), int(tz_hour), int(tz_min)
    if base is None:
        base = [1970, 1, 1, 0, 0]
    base = datetime(*base, tzinfo=timezone.utc).timestamp()
    ts = datetime(t_year, t_month, t_day, t_hour, t_min, t_seconds,
                  tzinfo=timezone(offset=timedelta(hours=tz_hour, minutes=tz_min))).timestamp()
    return ts - base


def merge_dicts(a: dict, b: dict):
    """
    a and b, two dictionary. Return updated a
        For example:
            a = {"1": 1, "2": {"1": 1, "2": 2, "3": 3, "5": {}, "6": [1, 2]}}
            b = {"1": 'b', "2": {"1": 'b', "2": 'b', "3": 'b', "4": 'b', "5": {"1": 'b'}, "6": [1, 2, 3]}}
            Will return {'1': 1, '2': {'1': 1, '2': 2, '3': 3, '5': {'1': 'b'}, '6': [1, 2], '4': 'b'}}
    """
    res = copy.deepcopy(a)
    for key, val in b.items():
        if key not in res.keys() and str(key).isnumeric():
            key = int(key)
        if key not in res.keys():
            res[key] = val
        elif isinstance(val, dict):
            res[key] = merge_dicts(res[key], val)
        else:
            continue
    return res


def update_dicts(a: dict, b: dict):
    """
    a and b, two dictionary. Return updated a
        For example:
            a = {"1": 1, "2": {"1": 1, "2": 2, "3": 3, "5": {}, "6": [1, 2]}}
            b = {"1": 'b', "2": {"1": 'b', "2": 'b', "3": 'b', "4": 'b', "5": {"1": 'b'}, "6": [1, 2, 3]}}
            Will return {'1': 'b', ...}
    """
    res = copy.deepcopy(a)
    for key, val in b.items():
        if key not in res.keys() and key.isnumeric():
            key = int(key)
        if key not in res.keys():
            res[key] = val
        elif isinstance(val, dict):
            res[key] = update_dicts(res[key], val)
        else:
            res[key] = val
    return res


def get_random_string(length: int = 8) -> str:
    return secrets.token_hex(max(int(length/2), 1))


def monte_carlo(func, inputs, confidence_level, **kwargs):
    """

    Parameters
    ----------
    func: Callable
    inputs: array
        two-dimensional array
    confidence_level: float
        [0, 1]

    Returns
    -------

    """
    N = len(inputs)
    l_range = int(0.5 * (1 - confidence_level) * N)
    r_range = N - l_range - 1
    values = np.transpose([func(*each, **kwargs) for each in inputs])
    cov = np.cov(values, rowvar=True)
    values = np.sort(values, axis=1)  # sort to find the lower and upper limitation at the given confidence level
    means = np.mean(values, axis=1).reshape(len(values), 1)
    limits = values[:, [l_range, r_range]]

    return np.concatenate((means, limits), axis=1), cov


def parser(v: str, t: type):
    if t == bool:
        if str(v).lower().rstrip() in ["true", "1"]:
            return True
    return t(v)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - files - basic
# ==========================================
#
#
#
"""

import os
import pickle
import json


def read(file_path):
    """ Read text files, default 'r', 'rb'
    Parameters
    ----------
    file_path

    Returns
    -------

    """
    try:
        with open(file_path, 'r') as f:
            params = json.load(f)
    except UnicodeDecodeError:
        with open(file_path, 'rb') as f:
            params = pickle.load(f)
    return params


def write(file_path, params):
    """
    Parameters
    ----------
    file_path
    params

    Returns
    -------

    """
    # with open(file_path, 'wb') as f:
    #     f.write(pickle.dumps(params))
    with open(file_path, 'w') as f:  # save serialized json data to a readable text
        f.write(json.dumps(params))
    return file_path


def delete(file_path):
    """
    Parameters
    ----------
    file_path

    Returns
    -------

    """
    try:
        os.remove(file_path)
    except Exception:
        return False
    else:
        return True



#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Show all Kfs age spectra
# ==========================================
#
#
# 
"""
import os
import numpy as np
import ararpy as ap
import pdf_maker as pm
import matplotlib
from matplotlib import cm
from matplotlib.collections import PathCollection

matplotlib.use('TkAgg')
matplotlib.rc('font',family='Arial', size=10)
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle


heating_schedule = {
    '20240621_24FY56a': 1,
    '20240630_24FY49a': 2,
    '20240705_24FY50a': 2,
    '20240710_24FY51a': 2,
    '20240714_24FY52': 2,
    '20240728_24FY55a': 2,
    '20240821_24FY67a': 2,
    '20240823_24FY71a': 2,
    '20240916_24FY92a': 2,
    '20240920_24FY88a': 2,
    '20240922_24FY89a': 2,
    '20240924_24FY70a': 2,
    '20241005_24FY72a': 3,
    '20241009_24FY62a': 3,
    '20241013_24FY64b': 3,
    '20241016_24FY66a': 4,
    '20241023_24FY68a': 5,
    '20241118_24FY93a': 7,
    '20241204_24FY53b': 7,
    '20241206_24FY54b': 6,
    '20241208_24FY87b': 6,
    '20241209_24FY86b': 6,
}

cc = {
    'Y49': 2,
    'Y50': 2,
    'Y51': 2,
    'Y52': 2,
    'Y53': 7,
    'Y54': 6,
    'Y55': 2,
    'Y56': 1,
    'Y62': 3,
    'Y64': 3,
    'Y66': 4,
    'Y67': 2,
    'Y68': 5,
    'Y70': 2,
    'Y71': 2,
    'Y72': 3,
    'Y86': 6,
    'Y87': 6,
    'Y88': 2,
    'Y89': 2,
    'Y93': 7,
    'Y92': 2,
}


def transform(ax: plt.Axes):
    xlabels = [i.get_text().replace('−', '-') for i in ax.get_xticklabels()]
    ylabels = [i.get_text().replace('−', '-') for i in ax.get_yticklabels()]
    linestyles = {'-': 'solid', '--': 'dashed', '-.': 'dashdot', ':': 'dotted'}

    series = []
    for i, line in enumerate(ax.lines):
        xy_data = line.get_xydata()  # [[x1, y1], [x2, y2], ...]
        line_style = linestyles.get(line.get_linestyle(), 'solid')
        series.append({
            'type': 'series.line', 'id': f'line-{i}', 'name': f'line-{i}',
            'color': line.get_color(), 'line_width': 1, 'line_style': line_style,
            'data': xy_data, 'line_caps': 'none'
        })
        if bool(line._marker):
            series.append({
                'type': 'series.scatter', 'id': f'line-marker-{i}', 'name': f'line-marker-{i}',
                'stroke_color': line.get_markeredgecolor(), 'fill_color': line.get_markerfacecolor(),
                'data': xy_data, 'size': 2,
                # 'symbol': line._marker.markers.get(line.get_marker(), 'square'),
                'symbol': 'rec'
            })
    for i, collection in enumerate(ax.collections):
        series.append({
            'type': 'series.scatter', 'id': f'scatter-{i}', 'name': f'{collection.get_label()}',
            'stroke_color': collection.get_edgecolor()[0][:3], 'fill_color': collection.get_edgecolor()[0][:3],
            'data': collection.get_offsets(), 'size': 2,
            'symbol': 'rec'
        })

    for i, text in enumerate(ax.texts):
        xy_data = text.get_position()  # [[x1, y1], [x2, y2], ...]
        series.append({
            'type': 'series.text', 'id': f'text-{i}', 'name': f'text-{i}',
            'color': text.get_color(), 'data': [xy_data], 'text': text.get_text().replace('\n', '<r>'),
            'size': 8
        })

    for child in ax._children:
        if isinstance(child, Rectangle):
            ld = [*child.get_xy(), child.get_width(), child.get_height()]
            series.append({
                'type': 'series.rect', 'id': f'rect-{child.get_gid()}', 'name': f'rect-{child.get_gid()}',
                'color': child.get_edgecolor()[:3], 'fill_color': child.get_facecolor()[:3], 'fill': True,
                'data': [ld], 'label': child.get_label(), 'clip': True
            })

    series.append({
        'type': 'series.text', 'id': f'title', 'name': f'title',
        'color': 'black', 'data': [[sum(ax.get_xlim()) / 2, ax.get_ylim()[1]]],
        'h_align': "middle", 'v_align': "top",
        'text': ax.get_title(), 'size': 8
    })

    if ax.legend_ is not None:
        for handle, text in zip(ax.legend_.legend_handles, ax.legend_.texts):
            series.append({
                'type': 'series.text', 'id': f'legend', 'name': f'legend',
                'color': text.get_color(), 'data': [[ax.get_xlim()[0], ax.get_ylim()[0]]],
                'h_align': "left", 'v_align': "bottom",
                'text': text.get_text(), 'size': 8
            })
            if isinstance(handle, plt.Line2D):
                series.append({
                    'type': 'series.line', 'id': f'legend-line', 'name': f'legend-line-{text.get_text()}',
                    'color': handle.get_color(), 'data':[[ax.get_xlim()[0], ax.get_ylim()[0]], [ax.get_xlim()[1], ax.get_ylim()[1]]],
                    'line_width': 1, 'line_style': linestyles.get(handle.get_linestyle(), 'solid')
                })
            if isinstance(handle, PathCollection):
                stroke_c = handle.get_edgecolor()[0][:3]
                stroke_c = f"#{int(stroke_c[0]*255):02x}{int(stroke_c[1]*255):02x}{int(stroke_c[2]*255):02x}"
                fill_c = handle.get_facecolor()[0][:3]
                fill_c = f"#{int(fill_c[0]*255):02x}{int(fill_c[1]*255):02x}{int(fill_c[2]*255):02x}"
                series.append({
                    'type': 'series.scatter', 'id': f'legend-scatter', 'name': f'legend-scatter-{text.get_text()}',
                    'stroke_color': stroke_c, 'fill_color': fill_c,
                    'data': [[sum(ax.get_xlim()) / 2, sum(ax.get_ylim()) / 2]],
                    'size': 2, 'symbol': 'rec'
                })

    data = {
        'xAxis': [{
            'extent': ax.get_xlim(), 'interval': xlabels, 'title': ax.get_xlabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'yAxis': [{
            'extent': ax.get_ylim(), 'interval': ylabels, 'title': ax.get_ylabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'series': series
    }

    # print(data)
    return data


def read_sample(arr_file_path):
    # 读取样品信息
    if not arr_file_path.endswith('.arr'):
        for root, dirs, files in os.walk(arr_file_path):
            for file in files:
                if file.endswith('.arr'):
                    arr_file_path = os.path.join(arr_file_path, file)
                    break
    sample = ap.from_arr(file_path=arr_file_path)
    name = sample.name()
    sequence = sample.sequence()
    nsteps = sequence.size
    te = np.array(sample.TotalParam[124], dtype=np.float64)
    ti = (np.array(sample.TotalParam[123], dtype=np.float64) / 60).round(2)  # time in minute
    nindex = {"40": 24, "39": 20, "38": 10, "37": 8, "36": 0}
    argon = "39"
    argon = "40"
    if argon in list(nindex.keys()):
        ar = np.array(sample.DegasValues[nindex[argon]], dtype=np.float64)  # 20-21 Argon
        sar = np.array(sample.DegasValues[nindex[argon] + 1], dtype=np.float64)
    elif argon == 'total':
        all_ar = np.array(sample.CorrectedValues, dtype=np.float64)  # 20-21 Argon
        ar, sar = ap.calc.arr.add(*all_ar.reshape(5, 2, len(all_ar[0])))
        ar = np.array(ar)
        sar = np.array(sar)
    else:
        raise KeyError
    age = np.array(sample.ApparentAgeValues[2], dtype=np.float64)  # 2-3 age
    sage = np.array(sample.ApparentAgeValues[3], dtype=np.float64)
    f = ar / ar.sum() * 100

    ar39 = np.array(sample.DegasValues[20], dtype=np.float64)
    ar40 = np.array(sample.DegasValues[24], dtype=np.float64)

    # f = np.ones(60) * 100 / 60

    x, y1, y2 = ap.calc.spectra.get_data(age, sage, f, cumulative=False)

    X = x + list(reversed(x))
    Y = y1 + list(reversed(y2))

    hs = heating_schedule[name]

    # if hs != 2:
    #     raise ValueError
    print(f"{name = }, {nsteps = }")

    # if max(y1[15:50]) > 50:
    #     print(name)

    only_te = []
    only_f = []
    for i, tei in enumerate(te):
        if tei not in only_te:
            only_te.append(tei)
            only_f.append(0)
        else:
            only_f[-1] += ar[i]

    age_plot = sample.AgeSpectraPlot



    # return np.cumsum(ar39) / sum(ar39), np.cumsum(ar40) / sum(ar40), name
    # return x, y1, name
    return X, Y, name
    # return list(range(nsteps+1)), np.insert(f, 0, 0), name
    # return only_te, np.cumsum(only_f) / np.sum(only_f), name
    # return te, f, name
    # return np.transpose([[i * 100 / 60 for i in range(60)], age - sage, [100 / 60 for i in range(60)], 2 * sage])


def check_name(name):
    samples = ['Y64', 'Y66', 'Y67', 'Y68', 'Y70', 'Y72', 'Y86', 'Y88', 'Y89']
    # samples = ['Y71', 'Y87', 'Y92', 'Y93']
    samples = ['Y49', 'Y55', 'Y56', 'Y62']
    samples = ['Y71', 'Y92', 'Y93', 'Y72', 'Y87']
    # samples = ['Y51', 'Y52', 'Y53', 'Y54', 'Y50']
    for i in samples:
        if i in name:
            return True
    raise ValueError

fig, axs = plt.subplots(1, 1, figsize=(6, 4))

axs.set_xlim(0, 20, auto=False)
axs.set_ylim(0, 10, auto=False)
axs.set_title(f'All Kfs age spectra', loc='center', y=1)
axs.set_ylabel(f'Apparent Age (Ma)')
axs.set_xlabel(f'Cumulative 39Ar released')

i = 0
v = cm.get_cmap('viridis')
colors = [v(i) for i in np.random.random(22)]


loc = r"C:\Users\Young\OneDrive\00-Projects\【2】个人项目\2024-10 Ar-Ar扩散理论\钾长石Arr"
for root, dirs, files in os.walk(loc):
    for file in files:
        if file.endswith('.arr'):
            # for smp in samples:
            #     if smp in file:
            arr_file_path = os.path.join(loc, file)
            try:
                check_name(file)
                x, y, name = read_sample(arr_file_path)
                # axs.plot(x, [yi + i * 10 for yi in y], label=name)
                axs.plot(x, y, label=name)
            except Exception as e:
                print(e)
                continue
            # continue
        i += 1

axs.legend(loc='best')
fig.tight_layout()
plt.show()


filename = f"39argon release excess ar - 2 - matplotlib"

params_list = {
    "page_size": 'a4', "ppi": 72, "width": 14, "height": 8,
    "pt_width": 0.8, "pt_height": 0.8, "pt_left": 0.16, "pt_bottom": 0.18,
    "offset_top": 0, "offset_right": 0, "offset_bottom": 20, "offset_left": 30,
    "plot_together": False, "show_frame": False,
    'xlabel_offset': 8, 'ylabel_offset': 2
}

plot_data = {
    "data": [
        transform(axs),
    ],
    "file_name": filename,
    "plot_names": [f"plotname"],
}

filepath = os.path.join(r"C:\Users\Young\Downloads", f"{filename}.pdf")
cvs = [[ap.smp.export.get_cv_from_dict(plot, **params_list) for plot in plot_data['data']]]
for i in range(len(cvs[0])):
    pt = cvs[0][i]._plot_areas[0]
    for index, legned in enumerate(list(filter(lambda cp: cp.name() == 'legend', pt._components))):
        legned._size = 7
        legned._z_index = 250
        legned._h_align = "left"
        legned._v_align = "center"
        if i == 0:  # age spectra
            legned._x = 175
            legned._y = 40 + index * 10
        elif i == 1:  # cooling history
            legned._x = 65
            legned._y = 135 - index * 10
        elif i == 2:  # arrhenius
            legned._x = 75
            legned._y = 40 + index * 10
        else:
            legned._x = 65
            legned._y = 40 + index * 10
        for comp in pt._components:
            if legned._text in comp.name() and "legend" in comp.name():
                comp._z_index = 250
                if isinstance(comp, pm.Scatter):
                    comp._x = legned._x - 10
                    comp._y = legned._y
                if isinstance(comp, pm.Line):
                    comp._start = [legned._x - 16, legned._y]
                    comp._end = [legned._x - 4, legned._y]


filepath = ap.smp.export.export_chart_to_pdf(cvs, filename, filepath)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - style
# ==========================================
#
#
#
"""

import traceback
import numpy as np
from .. import calc
from . import (sample as samples, basic, initial, plots, )

Sample = samples.Sample
Table = samples.Table
Plot = samples.Plot

TABLEHEADER = lambda sample_type, index: {
    'Unknown': ['',
                samples.SAMPLE_INTERCEPT_HEADERS, samples.BLANK_INTERCEPT_HEADERS,
                samples.CORRECTED_HEADERS, samples.DEGAS_HEADERS, samples.PUBLISH_TABLE_HEADERS,
                samples.SPECTRUM_TABLE_HEADERS, samples.ISOCHRON_TABLE_HEADERS,
                samples.TOTAL_PARAMS_HEADERS
                ],
    'Standard': ['',
                samples.SAMPLE_INTERCEPT_HEADERS, samples.BLANK_INTERCEPT_HEADERS,
                samples.CORRECTED_HEADERS, samples.DEGAS_HEADERS, samples.PUBLISH_TABLE_HEADERS_STD,
                samples.SPECTRUM_TABLE_HEADERS_STD, samples.ISOCHRON_TABLE_HEADERS,
                samples.TOTAL_PARAMS_HEADERS
                ],
    'Air': ['',
            samples.SAMPLE_INTERCEPT_HEADERS, samples.BLANK_INTERCEPT_HEADERS,
            samples.CORRECTED_HEADERS, samples.DEGAS_HEADERS, samples.PUBLISH_TABLE_HEADERS_AIR,
            samples.SPECTRUM_TABLE_HEADERS_AIR, samples.ISOCHRON_TABLE_HEADERS,
            samples.TOTAL_PARAMS_HEADERS
            ],
}[sample_type][index]


# =======================
# Reset plot style
# =======================
def set_plot_style(smp: Sample):
    """

    Parameters
    ----------
    smp

    Returns
    -------

    """
    # Reset styles
    initial.initial_plot_styles(smp, except_attrs=['data'])
    # Auto scale
    reset_plot_scale(smp)
    # Reset isochron data
    plots.reset_isochron_line_data(smp)
    # Auto position and contents of texts
    reset_text(smp)
    # Set title, which are deleted in initializing
    suffix = f"{smp.Info.sample.name} {smp.Info.sample.material}"
    for figure_id, figure in basic.get_components(smp).items():
        if isinstance(figure, Plot):
            if not hasattr(figure, 'title'):
                setattr(figure, 'title', Plot.Text())
            setattr(getattr(figure, 'title'), 'text', f"{suffix} {getattr(figure, 'name', '')}")


def reset_plot_scale(smp: Sample, only_figure: str = None):
    """
    Reset x- and y-axes scale
    Parameters
    ----------
    smp : sample instance
    only_figure

    Returns
    -------
    tuple of two tuples, (xscale, yscale)
    """
    for k, v in basic.get_components(smp).items():
        if not isinstance(v, Plot):
            continue
        if only_figure is not None and k != only_figure:
            continue
        if k == 'figure_1':
            try:
                # data = calc.arr.transpose(
                #     v.data + v.data + v.set1.data + v.set1.data + v.set2.data + v.set2.data)
                k0 = calc.arr.transpose(v.data)
                k1 = calc.arr.transpose(v.set1.data) if len(v.set1.data) != 0 else [[]] * 3
                k2 = calc.arr.transpose(v.set2.data) if len(v.set2.data) != 0 else [[]] * 3
                data = np.concatenate([k0, k1, k2], axis=1)
                ylist = np.concatenate([data[1], data[2]])
                yscale = calc.plot.get_axis_scale(ylist, min_interval=5, extra_count=1)
                xscale = [0, 100, 20]
            except (Exception, BaseException):
                print(traceback.format_exc())
                continue
        elif k == 'figure_3':
            try:
                xlist = v.data[0]
                xscale = calc.plot.get_axis_scale(xlist)
                yscale = [0, 0.004, 4, 0.001]
            except (Exception, BaseException):
                # print(traceback.format_exc())
                continue
        elif k == 'figure_7':
            try:
                xlist, ylist, zlist = v.data[0], v.data[2], v.data[4]
                xscale = calc.plot.get_axis_scale(xlist)
                yscale = calc.plot.get_axis_scale(ylist)
                zscale = calc.plot.get_axis_scale(zlist)
                setattr(getattr(v, 'zaxis'), 'min', zscale[0])
                setattr(getattr(v, 'zaxis'), 'max', zscale[1])
                setattr(getattr(v, 'zaxis'), 'split_number', zscale[2])
            except (Exception, BaseException):
                # print(traceback.format_exc())
                continue
        elif k == 'figure_9':
            try:
                xlist = v.set3.data[0] + [v.set3.data[0][i] - v.set3.data[1][i] for i in
                                          range(len(v.set3.data[0]))] + [
                            v.set3.data[0][i] + v.set3.data[1][i] for i in range(len(v.set3.data[0]))]
                ylist = v.set1.data[1]
                xscale = calc.plot.get_axis_scale(xlist)
                yscale = calc.plot.get_axis_scale(ylist)
            except (Exception, BaseException):
                print(traceback.format_exc())
                continue
        else:
            try:
                xlist = v.data[0]
                ylist = v.data[2]
                xscale = calc.plot.get_axis_scale(xlist)
                yscale = calc.plot.get_axis_scale(ylist)
            except (Exception, BaseException):
                # print(traceback.format_exc())
                continue
        setattr(getattr(v, 'xaxis', Plot.Axis()), 'min', xscale[0])
        setattr(getattr(v, 'xaxis', Plot.Axis()), 'max', xscale[1])
        setattr(getattr(v, 'xaxis', Plot.Axis()), 'split_number', xscale[2])
        setattr(getattr(v, 'yaxis', Plot.Axis()), 'min', yscale[0])
        setattr(getattr(v, 'yaxis', Plot.Axis()), 'max', yscale[1])
        setattr(getattr(v, 'yaxis', Plot.Axis()), 'split_number', yscale[2])

        if only_figure is not None and k == only_figure:
            return xscale, yscale


def reset_text(smp: Sample, only_figure: str = None):
    """
    Reset text position to default, if only figure is defined, only this figure will be reset.
    Parameters
    ----------
    smp
    only_figure

    Returns
    -------
    None
    """
    default_styles = initial.get_default_plot_style(smp)
    for figure_id in list(default_styles.keys()):
        if only_figure is not None and figure_id != only_figure:
            continue
        figure = basic.get_component_byid(smp, figure_id)
        text_1 = basic.get_plot_set(figure, 'Text for Set 1')
        if text_1 is not None:
            setattr(text_1, 'pos', [20, 40])
            setattr(text_1, 'text', "")
        text_2 = basic.get_plot_set(figure, 'Text for Set 2')
        if text_2 is not None:
            setattr(text_2, 'pos', [70, 40])
            setattr(text_2, 'text', "")


# =======================
# Reset plot style
# =======================
def set_table_style(sample: Sample):
    """

    Parameters
    ----------
    sample

    Returns
    -------

    """
    std = initial.initial(Sample())
    for key, comp in basic.get_components(sample).items():
        if isinstance(comp, Table):
            comp.header = TABLEHEADER(sample_type=sample.Info.sample.type, index=int(comp.id))
            comp.text_indexes = basic.get_component_byid(std, comp.id).text_indexes
            comp.set_coltypes()
            # comp.colcount = len(comp.header)
            # comp.coltypes = [{'type': 'numeric'}] * (comp.colcount)
            # text_indexes = getattr(comp, 'text_indexes', [0]) if hasattr(comp, 'text_indexes') else [0]
            # for i in text_indexes:
            #     comp.coltypes[i] = {'type': 'text'}





from . import (
    sample as samples, info, consts, json, basic, corr, raw, initial,
    plots, style, table, calculation, export, diffusion_funcs
)

Sample = samples.Sample
Table = samples.Table
Plot = samples.Plot
Set = samples.Plot.Set
Label = samples.Plot.Label
Axis = samples.Plot.Axis
Text = samples.Plot.Text
RawData = samples.RawData
Sequence = samples.Sequence
ArArBasic = samples.ArArBasic
ArArData = samples.ArArData

VERSION = samples.VERSION




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - files - xls
# ==========================================
#
#
#
"""
from xlrd import open_workbook, biffh


def open_xls(filepath: str):
    res = {}
    try:
        wb = open_workbook(filepath)
        for sheet_name in wb.sheet_names():
            sheet = wb.sheet_by_name(sheet_name)
            res[sheet_name] = [[sheet.cell(row, col).value for col in range(sheet.ncols)] for row in range(sheet.nrows)]
    except biffh.XLRDError as e:
        raise
    else:
        return res




#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Plot temperature calibration
# ==========================================
#
#
# 
"""
import datetime
import os
import re
import pytz

import numpy as np
import ararpy as ap
import pdf_maker as pm
import matplotlib
from matplotlib import cm
from matplotlib.collections import PathCollection

matplotlib.use('TkAgg')
matplotlib.rc('font',family='Arial', size=10)
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle


def transform(ax: plt.Axes):
    xlabels = [i.get_text().replace('−', '-') for i in ax.get_xticklabels()]
    ylabels = [i.get_text().replace('−', '-') for i in ax.get_yticklabels()]
    linestyles = {'-': 'solid', '--': 'dashed', '-.': 'dashdot', ':': 'dotted'}

    series = []
    for i, line in enumerate(ax.lines):
        xy_data = line.get_xydata()  # [[x1, y1], [x2, y2], ...]
        line_style = linestyles.get(line.get_linestyle(), 'solid')
        series.append({
            'type': 'series.line', 'id': f'line-{i}', 'name': f'line-{i}',
            'color': line.get_color(), 'line_width': 1, 'line_style': line_style,
            'data': xy_data, 'line_caps': 'none'
        })
        if bool(line._marker):
            series.append({
                'type': 'series.scatter', 'id': f'line-marker-{i}', 'name': f'line-marker-{i}',
                'stroke_color': line.get_markeredgecolor(), 'fill_color': line.get_markerfacecolor(),
                'data': xy_data, 'size': 2,
                # 'symbol': line._marker.markers.get(line.get_marker(), 'square'),
                'symbol': 'rec'
            })
    for i, collection in enumerate(ax.collections):
        series.append({
            'type': 'series.scatter', 'id': f'scatter-{i}', 'name': f'{collection.get_label()}',
            'stroke_color': collection.get_edgecolor()[0][:3], 'fill_color': collection.get_edgecolor()[0][:3],
            'data': collection.get_offsets(), 'size': 2,
            'symbol': 'rec'
        })

    for i, text in enumerate(ax.texts):
        xy_data = text.get_position()  # [[x1, y1], [x2, y2], ...]
        series.append({
            'type': 'series.text', 'id': f'text-{i}', 'name': f'text-{i}',
            'color': text.get_color(), 'data': [xy_data], 'text': text.get_text().replace('\n', '<r>'),
            'size': 8
        })

    for child in ax._children:
        if isinstance(child, Rectangle):
            ld = [*child.get_xy(), child.get_width(), child.get_height()]
            series.append({
                'type': 'series.rect', 'id': f'rect-{child.get_gid()}', 'name': f'rect-{child.get_gid()}',
                'color': child.get_edgecolor()[:3], 'fill_color': child.get_facecolor()[:3], 'fill': True,
                'data': [ld], 'label': child.get_label(), 'clip': True
            })

    series.append({
        'type': 'series.text', 'id': f'title', 'name': f'title',
        'color': 'black', 'data': [[sum(ax.get_xlim()) / 2, ax.get_ylim()[1]]],
        'h_align': "middle", 'v_align': "top",
        'text': ax.get_title(), 'size': 8
    })

    if ax.legend_ is not None:
        for handle, text in zip(ax.legend_.legend_handles, ax.legend_.texts):
            series.append({
                'type': 'series.text', 'id': f'legend', 'name': f'legend',
                'color': text.get_color(), 'data': [[ax.get_xlim()[0], ax.get_ylim()[0]]],
                'h_align': "left", 'v_align': "bottom",
                'text': text.get_text(), 'size': 8
            })
            if isinstance(handle, plt.Line2D):
                series.append({
                    'type': 'series.line', 'id': f'legend-line', 'name': f'legend-line-{text.get_text()}',
                    'color': handle.get_color(), 'data':[[ax.get_xlim()[0], ax.get_ylim()[0]], [ax.get_xlim()[1], ax.get_ylim()[1]]],
                    'line_width': 1, 'line_style': linestyles.get(handle.get_linestyle(), 'solid')
                })
            if isinstance(handle, PathCollection):
                stroke_c = handle.get_edgecolor()[0][:3]
                stroke_c = f"#{int(stroke_c[0]*255):02x}{int(stroke_c[1]*255):02x}{int(stroke_c[2]*255):02x}"
                fill_c = handle.get_facecolor()[0][:3]
                fill_c = f"#{int(fill_c[0]*255):02x}{int(fill_c[1]*255):02x}{int(fill_c[2]*255):02x}"
                series.append({
                    'type': 'series.scatter', 'id': f'legend-scatter', 'name': f'legend-scatter-{text.get_text()}',
                    'stroke_color': stroke_c, 'fill_color': fill_c,
                    'data': [[sum(ax.get_xlim()) / 2, sum(ax.get_ylim()) / 2]],
                    'size': 2, 'symbol': 'rec'
                })

    data = {
        'xAxis': [{
            'extent': ax.get_xlim(), 'interval': xlabels, 'title': ax.get_xlabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'yAxis': [{
            'extent': ax.get_ylim(), 'interval': ylabels, 'title': ax.get_ylabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'series': series
    }

    # print(data)
    return data

def read_libano(file_path):
    res = []

    if os.path.isdir(file_path):
        for root, dirs, files in os.walk(file_path):
            for file in files:
                if "libano" in file:
                    file_path = os.path.join(file_path, file)
                    break
    print(f"{file_path = }")
    if os.path.exists(file_path):
        for line in open(file_path, "r"):
            temp = re.findall(r"Z;\d+;\d+", line)
            setpoint, reading = re.findall(r"\d+", temp[0])
            dt_object = datetime.datetime.strptime(
                re.findall(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", line)[0],
                "%Y-%m-%dT%H:%M:%SZ")

            timezone = pytz.utc
            dt_object = timezone.localize(dt_object)

            res.append([dt_object, int(setpoint), int(reading)])
            # res[int(dt_object.timestamp())] = {
            #     "set": int(setpoint), "read": int(reading)
            # }

    return res


def read_test_log(file_path):
    # res = {}
    res = []

    if os.path.isdir(file_path):
        for root, dirs, files in os.walk(file_path):
            for file in files:
                if "Test" in file:
                    file_path = os.path.join(file_path, file)
                    break

    print(f"{file_path = }")
    if os.path.exists(file_path):
        for line in open(file_path, "r"):
            if len(line) <= 20:
                continue
            split = line.split(";")
            time_str = split[0]
            temp_str = split[1]

            dt_object = datetime.datetime.strptime(
                re.findall(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", time_str)[0],
                "%Y-%m-%dT%H:%M:%SZ")

            timezone = pytz.utc
            dt_object = timezone.localize(dt_object)

            # res[int(dt_object.timestamp())] = {
            #     "read": int(temp_str)
            # }
            res.append([dt_object, int(temp_str)])

    return res


loc = r"C:\Users\Young\OneDrive\Documents\Libano Data\2024-03-13"

s = r"202403131203-libano.log"

filepath = os.path.join(loc, s)

outside = read_libano(loc)
inside = read_test_log(loc)


fig, axs = plt.subplots(1, 1, figsize=(6, 4))

# axs.set_xlim(0, 20, auto=True)
# axs.set_ylim(0, 10, auto=True)
# axs.set_title(f'Temperature check', loc='center', y=1)
# axs.set_ylabel(f'Time (ms)')
# axs.set_xlabel(f'Temperature (°C)')

i = 0
v = cm.get_cmap('viridis')
colors = [v(i) for i in np.random.random(22)]

x1, y1_set, y1_read = ap.calc.arr.transpose(outside)
x2, y2_read = ap.calc.arr.transpose(inside)

print(x1)
print(y1_set)


axs.plot(x1, y1_set, label="Setpoints", c="green")
axs.plot(x1, y1_read, label="Outside readings", c="blue")
axs.plot(x2, y2_read, label="Inside readings", c="red")

# 设置x轴格式
axs.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M:%S'))
axs.xaxis.set_major_locator(mdates.SecondLocator(interval=7200))  # 每分钟一个主刻度
axs.xaxis.set_minor_locator(mdates.SecondLocator(interval=3600))   # 每10秒一个次刻度
axs.set_xlim(
    datetime.datetime.strptime("2024-03-13T15:40:00Z", "%Y-%m-%dT%H:%M:%SZ"),
    datetime.datetime.strptime("2024-03-13T16:20:00Z", "%Y-%m-%dT%H:%M:%SZ")
)
axs.set_ylim(650, 800)

# axs.legend(loc='best')
fig.tight_layout()
plt.show()

#
#
# filename = f"39argon release excess ar - 2 - matplotlib"
#
# params_list = {
#     "page_size": 'a4', "ppi": 72, "width": 14, "height": 8,
#     "pt_width": 0.8, "pt_height": 0.8, "pt_left": 0.16, "pt_bottom": 0.18,
#     "offset_top": 0, "offset_right": 0, "offset_bottom": 20, "offset_left": 30,
#     "plot_together": False, "show_frame": False,
#     'xlabel_offset': 8, 'ylabel_offset': 2
# }
#
# plot_data = {
#     "data": [
#         transform(axs),
#     ],
#     "file_name": filename,
#     "plot_names": [f"plotname"],
# }
#
# filepath = os.path.join(r"C:\Users\Young\Downloads", f"ceshi.pdf")
# cvs = [[ap.smp.export.get_cv_from_dict(plot, **params_list) for plot in plot_data['data']]]
# for i in range(len(cvs[0])):
#     pt = cvs[0][i]._plot_areas[0]
#     for index, legned in enumerate(list(filter(lambda cp: cp.name() == 'legend', pt._components))):
#         legned._size = 7
#         legned._z_index = 250
#         legned._h_align = "left"
#         legned._v_align = "center"
#         if i == 0:  # age spectra
#             legned._x = 175
#             legned._y = 40 + index * 10
#         elif i == 1:  # cooling history
#             legned._x = 65
#             legned._y = 135 - index * 10
#         elif i == 2:  # arrhenius
#             legned._x = 75
#             legned._y = 40 + index * 10
#         else:
#             legned._x = 65
#             legned._y = 40 + index * 10
#         for comp in pt._components:
#             if legned._text in comp.name() and "legend" in comp.name():
#                 comp._z_index = 250
#                 if isinstance(comp, pm.Scatter):
#                     comp._x = legned._x - 10
#                     comp._y = legned._y
#                 if isinstance(comp, pm.Line):
#                     comp._start = [legned._x - 16, legned._y]
#                     comp._end = [legned._x - 4, legned._y]
#
#
# filepath = ap.smp.export.export_chart_to_pdf(cvs, filename, filepath)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - plots
# ==========================================
#
#
#
"""

import traceback
import numpy as np

from scipy.signal import find_peaks


from .. import calc
from .sample import Sample, Table, Plot
from . import basic, initial

Set = Plot.Set
Label = Plot.Label
Axis = Plot.Axis
Text = Plot.Text


ISOCHRON_INDEX_DICT = {
    'figure_2': {'data_index': [0, 5], 'name': 'Normal Isochron', 'figure_type': 1},
    'figure_3': {'data_index': [6, 11], 'name': 'Inverse Isochron', 'figure_type': 2},
    'figure_4': {'data_index': [12, 17], 'name': 'Cl Correlation 1', 'figure_type': 1},
    'figure_5': {'data_index': [18, 23], 'name': 'Cl Correlation 2', 'figure_type': 2},
    'figure_6': {'data_index': [24, 29], 'name': 'Cl Correlation 3', 'figure_type': 3},
    'figure_7': {'data_index': [30, 39], 'name': 'Cl Correlation 3D', 'figure_type': 0},
}


# =======================
# Reset plot data
# =======================
def set_plot_data(sample: Sample, isInit: bool = True, isIsochron: bool = True,
                  isPlateau: bool = True, **kwargs):
    """
    Parameters
    ----------
    sample
    isInit
    isIsochron
    isPlateau
    kwargs

    Returns
    -------

    """
    # print(f"isInit: {isInit}, isIsochron: {isIsochron}, isPlateau: {isPlateau}")

    # Initialization, apply age spectra data and isochron plot data
    if isInit:
        try:
            initial_plot_data(sample)
        except (Exception, BaseException):
            print("initial_plot_data(sample) error:\n", traceback.format_exc())
            pass

    # Recalculate isochron lines
    if isIsochron:
        try:
            recalc_isochrons(sample, **kwargs)
        except (Exception, BaseException):
            print("recalc_isochrons(sample, **kwargs) error:\n", traceback.format_exc())
            pass
        try:
            reset_isochron_line_data(sample)
        except (Exception, BaseException):
            print("reset_isochron_line_data(sample):\n", traceback.format_exc())
            pass

    # Recalculate plateaus
    if isPlateau:
        try:
            recalc_plateaus(sample)
        except (Exception, BaseException):
            print("recalc_plateaus(sample) error:\n", traceback.format_exc())
            pass


# =======================
# Initialize plot data
# =======================
def initial_plot_data(smp: Sample):
    """
    Assign initial data for plots
    Parameters
    ----------
    smp : Sample instance

    Returns
    -------
    None
    """
    for key, val in ISOCHRON_INDEX_DICT.items():
        figure = basic.get_component_byid(smp, key)
        try:
            # data = [x, sx, y, sy, (z, sz,) r1, (r2, r3,), index from 1]
            figure.data = smp.IsochronValues[val['data_index'][0]:val['data_index'][1]] + \
                          [[i + 1 for i in range(len(smp.SequenceName))]]
        except (Exception, BaseException):
            print(traceback.format_exc())
            figure.data = [[]] * (val['data_index'][1] - val['data_index'][0]) + \
                          [[i + 1 for i in range(len(smp.SequenceName))]]
        finally:
            # Ellipse
            if key != 'figure_7':
                ellipse_data = []
                for point in calc.arr.transpose(figure.data[:5]):
                    if '' not in point and None not in point:
                        ellipse_data.append(calc.isochron.get_ellipse(*point))
                        getattr(figure, 'ellipse', Set(id='ellipse')).data = ellipse_data

    # Set age spectra lines
    # Try to calculate total gas age
    try:
        if str(smp.Info.sample.type).lower() == "unknown":
            a0, e0 = sum(smp.DegasValues[24]), pow(sum([i ** 2 for i in smp.DegasValues[25]]), 0.5)
            a1, e1 = sum(smp.DegasValues[20]), pow(sum([i ** 2 for i in smp.DegasValues[21]]), 0.5)
            handle = basic.calc_age
        elif str(smp.Info.sample.type).lower() == "standard":
            a0, e0 = sum(smp.DegasValues[24]), pow(sum([i ** 2 for i in smp.DegasValues[25]]), 0.5)
            a1, e1 = sum(smp.DegasValues[20]), pow(sum([i ** 2 for i in smp.DegasValues[21]]), 0.5)
            handle = basic.calc_j
        elif str(smp.Info.sample.type).lower() == "air":
            a0, e0 = sum(smp.DegasValues[26]), pow(sum([i ** 2 for i in smp.DegasValues[27]]), 0.5)
            a1, e1 = sum(smp.DegasValues[ 0]), pow(sum([i ** 2 for i in smp.DegasValues[ 1]]), 0.5)
            handle = basic.calc_mdf
        else:
            raise TypeError(f"Sample type is not supported: {smp.Info.sample.type}")
        total_f = [a0 / a1, calc.err.div((a0, e0), (a1, e1))]
        total_age = handle(total_f[:2], smp=smp)
    except (Exception, BaseException):
        print(traceback.format_exc())
        total_f = [np.nan] * 2
        total_age = [np.nan] * 4
    smp.Info.results.age_spectra['TGA'].update(
        {'Ar39': 100, 'F': total_f[0], 'sF': total_f[1], 'age': total_age[0],
         's1': total_age[1], 's2': total_age[2], 's3': total_age[3], 'Num': len(smp.DegasValues[24])}
    )
    try:
        smp.AgeSpectraPlot.data = calc.spectra.get_data(*smp.ApparentAgeValues[2:4], smp.ApparentAgeValues[7])
        smp.AgeSpectraPlot.data = calc.arr.transpose(smp.AgeSpectraPlot.data)
    except (Exception, BaseException):
        print(traceback.format_exc())
        smp.AgeSpectraPlot.data = []

    # Degassing plot
    try:
        if not hasattr(smp, 'DegasPatternPlot'):
            setattr(smp, 'DegasPatternPlot', Plot(id='figure_8', name='Degas Pattern Plot'))
        isotope_percentage = lambda l: [e / sum(l) * 100 if sum(l) != 0 else 0 for e in l]
        smp.DegasPatternPlot.data = [
            isotope_percentage(smp.DegasValues[0]),  # Ar36a
            isotope_percentage(smp.DegasValues[8]),  # Ar37Ca
            isotope_percentage(smp.DegasValues[10]),  # Ar38Cl
            isotope_percentage(smp.DegasValues[20]),  # Ar39K
            isotope_percentage(smp.DegasValues[24]),  # Ar40r
            isotope_percentage(smp.CorrectedValues[0]),  # Ar36
            isotope_percentage(smp.CorrectedValues[2]),  # Ar37
            isotope_percentage(smp.CorrectedValues[4]),  # Ar38
            isotope_percentage(smp.CorrectedValues[6]),  # Ar39
            isotope_percentage(smp.CorrectedValues[8]),  # Ar40
        ]
        smp.DegasPatternPlot.info = [True] * 10
    except Exception as e:
        print(traceback.format_exc())
        pass

    # Set age distribution plot data
    try:
        recalc_agedistribution(smp)
    except Exception as e:
        print(traceback.format_exc())


# =======================
# Isochron results
# =======================
def recalc_isochrons(sample: Sample, **kwargs):
    """

    Parameters
    ----------
    sample
    kwargs

    Returns
    -------

    """
    figures = kwargs.pop('figures', ['figure_2', 'figure_3', 'figure_4', 'figure_5', 'figure_6', 'figure_7', ])
    for key, val in ISOCHRON_INDEX_DICT.items():
        if key not in figures:
            continue
        figure = basic.get_component_byid(sample, key)
        figure.set3.data, figure.set1.data, figure.set2.data = \
            sample.UnselectedSequence.copy(), sample.SelectedSequence1.copy(), sample.SelectedSequence2.copy()
        for index, sequence in enumerate([figure.set1.data, figure.set2.data, figure.set3.data]):
            set_data = calc.arr.partial(
                sample.IsochronValues, rows=sequence, cols=list(range(*val['data_index'])))
            if key != 'figure_7':
                iso_res = get_isochron_results(
                    set_data, figure_type=val["figure_type"], smp=sample, sequence=sequence)
                sample.Info.results.isochron[figure.id].update({index: iso_res})
            else:
                iso_res = get_3D_results(data=set_data, sequence=sequence, sample=sample)
                sample.Info.results.isochron[figure.id].update({index: iso_res})


def get_isochron_results(data: list, smp: Sample, sequence, figure_type: int = 0):
    """
    Get isochron figure results based on figure type.
    Parameters
    ----------
    data : isochron figure data, 5 columns list
    smp : sample instance
    sequence : data section index
    figure_type : int, 0 for normal isochron, 1 for inverse isochron, 2 for K-Cl-Ar plot 3

    Returns
    -------
    dict, isochron regression results, keys are [
        'k', 'sk', 'm1', 'sm1', 'MSWD', 'abs_conv', 'iter', 'mag', 'R2', 'Chisq', 'Pvalue',
        'rs',  'age', 's1', 's2', 's3', 'conv', 'initial', 'sinitial', 'F', 'sF'
    ]
    """
    reg_res_index = [
        'k', 'sk', 'm1', 'sm1',
        'MSWD', 'abs_conv', 'iter', 'mag', 'R2', 'Chisq', 'Pvalue',
        'rs',  # 'rs' means relative error of the total sum
        'cov_b_m'
    ]
    age_res_index = ['age', 's1', 's2', 's3', ]
    iso_res = dict(zip(
        [*reg_res_index, *age_res_index, 'conv', 'initial', 'sinitial', 'F', 'sF'],
        [np.nan] * (len(reg_res_index + age_res_index) + 5)
    ))

    if len(sequence) < 3:
        return iso_res
    try:
        regression_method = {
            "york-2": calc.regression.york2, "olst": calc.regression.olst
        }.get(smp.TotalParam[97][min(sequence)].lower(), calc.regression.york2)
        regression_res = regression_method(*data[:5])
    except (Exception, BaseException):
        # print(f"Warning: {traceback.format_exc()}")
        return iso_res
    else:
        iso_res.update(dict(zip(reg_res_index, regression_res)))
        if figure_type == 1:
            iso_res.update(zip(['initial', 'sinitial'], regression_res[0:2]))
            iso_res.update(zip(['F', 'sF'], regression_res[2:4]))
        elif figure_type == 2:
            iso_res.update(zip(['initial', 'sinitial'], calc.arr.rec(regression_res[0:2])))
            k = regression_method(*data[2:4], *data[0:2], data[4])
            iso_res.update(zip(['F', 'sF'], calc.arr.rec(k[0:2])))
        elif figure_type == 3:
            iso_res.update(zip(['initial', 'sinitial'], regression_res[2:4]))
            iso_res.update(zip(['F', 'sF'], regression_res[0:2]))
    # age, analytical err, internal err, full external err
    try:
        age = basic.calc_age([iso_res['F'], iso_res['sF']], smp=smp)
        iso_res.update(dict(zip(age_res_index, age)))
    except ValueError:
        pass
    return iso_res


def get_3D_results(data: list, sequence: list, sample: Sample):
    """
    Get 3D regression results.
    Parameters
    ----------
    data : 3D regression data with 9 columns.
    sequence : list, data section index
    sample : sample instance

    Returns
    -------
    dict, isochron regression results, with keys = [
        'k', 'sk', 'm1', 'sm1', 'm2', 'sm2',
        'S', 'MSWD', 'R2', 'abs_conv', 'iter', 'mag', 'Chisq', 'Pvalue',
        'rs', 'age', 's1', 's2', 's3', 'conv', 'initial', 'sinitial', 'p_Cl', 'F', 'sF'
    ]
    """
    reg_res_index = [
        'k', 'sk', 'm1', 'sm1', 'm2', 'sm2',
        'S', 'MSWD', 'R2', 'abs_conv', 'iter', 'mag', 'Chisq', 'Pvalue',
        'rs',  # 'rs' means relative error of the total sum
    ]
    age_res_index = ['age', 's1', 's2', 's3', ]
    iso_res = dict(zip(
        [*reg_res_index, *age_res_index,
         'conv', 'initial', 'sinitial', 'p_Cl', 'F', 'sF'],
        [np.nan] * (len(reg_res_index + age_res_index) + 8)
    ))
    try:
        if len(sequence) < 4:
            raise ValueError(f"Data points not enough.")
        k = calc.regression.wtd_3D_regression(*data[:9])
        ar38ar36 = sample.TotalParam[4][0]
        sar38ar36 = sample.TotalParam[5][0] * sample.TotalParam[4][0] / 100
        ar40ar36 = (k[2] + k[4] * ar38ar36) * -1 / k[0]
        sar40ar36 = calc.err.div(
            ((k[2] + k[4] * ar38ar36) * -1,
             calc.err.add(k[3], calc.err.mul((k[4], k[5]), (ar38ar36, sar38ar36)))), (k[0], k[1]))
        f = 1 / k[0]
        sf = calc.err.div((1, 0), (k[0], k[1]))
        try:
            PQ = -1 * k[4] / k[2]
            Q = 1 - np.exp(-1 * sample.TotalParam[46][0] * sum(sample.TotalParam[32]) / len(sample.TotalParam[32]))
            P = PQ / Q
        except:
            # print(f"Warning: {traceback.format_exc()}")
            P = 0
        age = basic.calc_age([f, sf], smp=sample)
    except:
        # print(f"Warning: {traceback.format_exc()}")
        pass
    else:
        iso_res.update(dict(zip(iso_res, [*k, *age, np.nan, ar40ar36, sar40ar36, P, f, sf])))
    return iso_res


def reset_isochron_line_data(smp: Sample):
    """
    Set isochron regression lines
    Parameters
    ----------
    smp : sample instance

    Returns
    -------
    None, set regression lines data to sample instance.
    """
    for k, v in basic.get_components(smp).items():
        if not isinstance(v, Plot):
            continue
        for index in [0, 1]:
            try:
                xscale, yscale = [v.xaxis.min, v.xaxis.max], [v.yaxis.min, v.yaxis.max]
                coeffs = [smp.Info.results.isochron[k][index]['k'], smp.Info.results.isochron[k][index]['m1']]
                line_point = calc.isochron.get_line_points(xscale, yscale, coeffs)
                setattr(getattr(v, ['line1', 'line2'][index]), 'data', line_point)
                setattr(getattr(v, ['text1', 'text2'][index]), 'text', "")  # 注意和js的配合，js那边根据text是否为空判断是否重新生成文字
            except Exception:
                # print(traceback.format_exc())
                continue


def set_selection(smp: Sample, index: int, mark: int):
    """
    Parameters
    ----------
    smp : sample instance
    index : int, data point index
    mark : int, 0 for unselected, 1 for set1, 2 for set2

    Returns
    -------

    """
    if mark not in [1, 2, '1', '2']:
        raise ValueError(f"{mark = }, mark must be 1 or 2.")

    def seq(_i): return [smp.UnselectedSequence, smp.SelectedSequence1, smp.SelectedSequence2][_i]

    if index in seq(mark):
        seq(mark).remove(index)
        smp.UnselectedSequence.append(index)
    else:
        for i in [0, [0, 2, 1][mark]]:
            if index in seq(i):
                seq(i).remove(index)
        seq(mark).append(index)
    smp.IsochronMark = [
        '1' if i in smp.SelectedSequence1 else '2' if i in smp.SelectedSequence2 else '' for i in
        range(len(smp.IsochronValues[2]))]
    #
    smp.Info.results.selection[0]['data'] = smp.SelectedSequence1
    smp.Info.results.selection[1]['data'] = smp.SelectedSequence2
    smp.Info.results.selection[2]['data'] = smp.UnselectedSequence


# =======================
# Age spectra results
# =======================
def recalc_plateaus(sample: Sample, **kwargs):
    if sample.Info.sample.type == "Unknown":
        return recalc_age_plateaus(sample, **kwargs)
    if sample.Info.sample.type == "Standard":
        return recalc_j_plateaus(sample, **kwargs)
    if sample.Info.sample.type == "Air":
        return recalc_mdf_plateaus(sample, **kwargs)


def recalc_age_plateaus(sample: Sample, **kwargs):
    """
    Calculate age plateaus results
    Parameters
    ----------
    sample : sample instance
    kwargs : optional args, keys in [r1, sr1, r2, sr2]

    Returns
    -------
    None
    """
    ar39k, sar39k = calc.arr.mul(sample.DegasValues[20:22], sample.TotalParam[136:138])
    ar40r, sar40r = sample.DegasValues[24:26]
    ar40rar39k = calc.arr.div([ar40r, sar40r], [ar39k, sar39k])
    params_initial_ratio = calc.arr.partial(sample.TotalParam, cols=list(range(115, 120)))
    ratio_set1 = [[], []]
    ratio_set2 = [[], []]
    for row, item in enumerate(params_initial_ratio[0]):
        if str(item) == '0':
            ratio_set1[0].append(sample.Info.results.isochron['figure_3'][0]['initial'])
            ratio_set1[1].append(sample.Info.results.isochron['figure_3'][0]['sinitial'])
            ratio_set2[0].append(sample.Info.results.isochron['figure_3'][1]['initial'])
            ratio_set2[1].append(sample.Info.results.isochron['figure_3'][1]['sinitial'])
        elif str(item) == '1':
            ratio_set1[0].append(sample.Info.results.isochron['figure_2'][0]['initial'])
            ratio_set1[1].append(sample.Info.results.isochron['figure_2'][0]['sinitial'])
            ratio_set2[0].append(sample.Info.results.isochron['figure_2'][1]['initial'])
            ratio_set2[1].append(sample.Info.results.isochron['figure_2'][1]['sinitial'])
        elif str(item) == '2':
            ratio_set1[0].append(params_initial_ratio[1][row])
            ratio_set1[1].append(params_initial_ratio[2][row])
            ratio_set2[0].append(params_initial_ratio[3][row])
            ratio_set2[1].append(params_initial_ratio[4][row])
        else:
            ratio_set1[0].append(298.56)
            ratio_set1[1].append(0.31)
            ratio_set2[0].append(298.56)
            ratio_set2[1].append(0.31)

    # Weighted mean ages of two sets with the given initial ratios
    try:
        set1_res, set1_age, set1_data = get_plateau_results(
            sample, sample.SelectedSequence1, calc_ar40ar39(*ratio_set1, smp=sample),
            ar39k_percentage=np.array(ar39k) / np.sum(ar39k),
            **kwargs)
    except ValueError:
        # print(traceback.format_exc())
        pass
        # raise ValueError(f"Set 1 Plateau results calculation error.")
    else:
        sample.Info.results.age_plateau.update({0: set1_res})
        sample.AgeSpectraPlot.set1.data = calc.arr.transpose(set1_data)
        sample.AgeSpectraPlot.text1.text = ""  # 注意和js的配合，js那边根据text是否为空判断是否重新生成文字
    try:
        set2_res, set2_age, set2_data = get_plateau_results(
            sample, sample.SelectedSequence2, calc_ar40ar39(*ratio_set2, smp=sample),
            ar39k_percentage=np.array(ar39k) / np.sum(ar39k),
            **kwargs)
    except ValueError:
        # print(traceback.format_exc())
        pass
        # raise ValueError(f"Set 2 Plateau results calculation error.")
    else:
        sample.Info.results.age_plateau.update({1: set2_res})
        sample.AgeSpectraPlot.set2.data = calc.arr.transpose(set2_data)
        sample.AgeSpectraPlot.text2.text = ""  # 注意和js的配合，js那边根据text是否为空判断是否重新生成文字

    # Weighted mean ages of two sets with 298.56 (defoult air ratio)
    try:
        set1_res = get_wma_results(
            sample, sample.SelectedSequence1, ar40rar39k=ar40rar39k, ar39k_percentage=np.array(ar39k) / np.sum(ar39k))
    except ValueError:
        pass
        # raise ValueError(f"Set 1 WMA calculation error.")
    else:
        sample.Info.results.age_spectra.update({0: set1_res})
    try:
        set2_res = get_wma_results(
            sample, sample.SelectedSequence2, ar40rar39k=ar40rar39k, ar39k_percentage=np.array(ar39k) / np.sum(ar39k))
    except ValueError:
        pass
        # raise ValueError(f"Set 2 WMA calculation error.")
    else:
        sample.Info.results.age_spectra.update({1: set2_res})

    # # """3D corrected plateaus"""
    # # 3D ratio, 36Ar(a+cl)/40Ar(a+r), 38Ar(a+cl)/40Ar(a+r), 39Ar(k)/40Ar(a+r),
    # ar40ar = calc_funcs.list_sub(*sample.CorrectedValues[8:10], *sample.DegasValues[30:32])
    # # 36Ar deduct Ca, that is sum of 36Ara and 36ArCl
    # ar36acl = calc_funcs.list_sub(*sample.CorrectedValues[0:2], *sample.DegasValues[4:6])
    # # 38Ar deduct K and Ca, that is sum of 38Ara and 38ArCl
    # ar38acl = calc_funcs.list_sub(*calc_funcs.list_sub(*sample.CorrectedValues[4:6], *sample.DegasValues[16:18]),
    #                               *sample.DegasValues[18:20])
    # # 39ArK
    # ar39k = sample.DegasValues[20:22]
    #
    # # 40Arr
    # def get_modified_f(c, sc, a, sa, b, sb):
    #     ar40r = list(map(lambda zi, xi, yi: zi - a * xi - b * yi, ar40ar[0], ar36acl[0], ar38acl[0]))
    #     sar40r = list(map(lambda zi, szi, xi, sxi, yi, syi:
    #                       calc.err.add(szi, calc_funcs.error_mul((xi, sxi), (a, sa)),
    #                                            calc_funcs.error_mul((yi, syi), (b, sb))),
    #                       *ar40ar, *ar36acl, *ar38acl))
    #     f = list(map(lambda ar40ri, ar39ki: ar40ri / ar39ki, ar40r, ar39k[0]))
    #     sf = list(map(lambda ar40ri, sar40ri, ar39ki, sar39ki:
    #                   calc.err.div((ar40ri, sar40ri), (ar39ki, sar39ki)),
    #                   ar40r, sar40r, *ar39k))
    #     return [f, sf]
    #
    # isochron_7 = calc_funcs.get_3D_isochron(*ar36acl, *ar38acl, *ar40ar, *ar39k)
    # [set1_data, set2_data, set3_data] = basic_funcs.getIsochronSetData(
    #     isochron_7, sample.SelectedSequence1, sample.SelectedSequence2, sample.UnselectedSequence)
    #
    # __isochron_7 = calc_funcs.get_3D_isochron(*ar36acl, *ar38acl, *ar39k, *ar40ar)
    # [__set1_data, __set2_data, __set3_data] = basic_funcs.getIsochronSetData(
    #     __isochron_7, sample.SelectedSequence1, sample.SelectedSequence2, sample.UnselectedSequence)
    #
    # def __get_modified_f(c, sc, a, sa, b, sb):
    #     f = list(
    #         map(lambda zi, xi, yi: 1 / (zi - a * xi - b * yi), __isochron_7[4], __isochron_7[0], __isochron_7[2]))
    #     sf = [0] * len(f)
    #     return [f, sf]
    #
    # # set 1:
    # try:
    #     k = calc_funcs.wtd_3D_regression(*set1_data[:9])
    #     set1_ar40rar39k = get_modified_f(*k[:6])
    #
    #     # __k = calc_funcs.wtd_3D_regression(*__set1_data[:9])
    #     # __set1_ar40rar39k = __get_modified_f(*__k[:6])
    #     #
    #     # for i in range(len(set1_ar40rar39k[0])):
    #     #     print(f"{set1_ar40rar39k[0][i]} == {__set1_ar40rar39k[0][i]}")
    #     #
    #     # k = calc_funcs.wtd_3D_regression(*__set1_data[:9])
    #     # set1_ar40rar39k = __get_modified_f(*k[:6])
    #
    # except:
    #     print(traceback.format_exc())
    #     set1_ar40rar39k = [[0] * len(ar39k[0]), [0] * len(ar39k[0])]
    # # set 2:
    # try:
    #     k = calc_funcs.wtd_3D_regression(*set2_data[:9])
    #     set2_ar40rar39k = get_modified_f(*k[:6])
    # except:
    #     set2_ar40rar39k = [[0] * len(ar39k[0]), [0] * len(ar39k[0])]
    # set4_age, set4_data, set4_wmf, set4_wmage, set4_text = \
    #     get_plateau_results(sample, sample.SelectedSequence1, set1_ar40rar39k)
    # set5_age, set5_data, set5_wmf, set5_wmage, set5_text = \
    #     get_plateau_results(sample, sample.SelectedSequence2, set2_ar40rar39k)
    # # Set set4 and set5
    # sample.AgeSpectraPlot.set4.data = calc.arr.transpose(set4_data)
    # sample.AgeSpectraPlot.set5.data = calc.arr.transpose(set5_data)
    # sample.AgeSpectraPlot.set4.info = [*set4_wmf, *set4_wmage]  # Info = weighted mean f, sf, np, mswd, age, s, s, s
    # sample.AgeSpectraPlot.set5.info = [*set5_wmf, *set5_wmage]  # Info = weighted mean f, sf, np, mswd, age, s, s, s
    # # """end"""


def recalc_mdf_plateaus(sample: Sample, **kwargs):
    """
    Calculate age plateaus results
    Parameters
    ----------
    sample : sample instance
    kwargs : optional args, keys in [r1, sr1, r2, sr2]

    Returns
    -------
    None
    """
    ar36a, sar36a = calc.arr.mul(sample.DegasValues[0:2], sample.TotalParam[136:138])
    ar40aar36a = sample.ApparentAgeValues[0:2]
    mdf = sample.ApparentAgeValues[2:4]

    try:
        set1_res, _, set1_data = get_plateau_results(sample, sample.SelectedSequence1, ar40rar39k=ar40aar36a)
    except ValueError:
        pass
    else:
        sample.Info.results.age_plateau.update({0: set1_res})
        sample.AgeSpectraPlot.set1.data = calc.arr.transpose(set1_data)
        sample.AgeSpectraPlot.text1.text = ""

    try:
        set2_res, _, set2_data = get_plateau_results(sample, sample.SelectedSequence2, ar40rar39k=ar40aar36a)
    except ValueError:
        pass
    else:
        sample.Info.results.age_plateau.update({1: set2_res})
        sample.AgeSpectraPlot.set2.data = calc.arr.transpose(set2_data)
        sample.AgeSpectraPlot.text2.text = ""


def calc_ar40ar39(r, sr, smp):
    """
    Calculate Ar40r / Ar39K based on passed initial ratio.
    Parameters
    ----------
    r : ratio value, float or list
    sr : error of the ratio, same type as r
    smp : sample instance

    Returns
    -------
    Two dimensional list, Ar40r / Ar39K values and errors
    """
    try:
        ar36a = np.array(smp.DegasValues[0:2])
        ar39k = calc.arr.mul(smp.DegasValues[20:22], smp.TotalParam[136:138])
        ar40 = smp.CorrectedValues[8:10]
        ar40k = smp.DegasValues[30:32]
        size = ar36a.shape[-1]
        if isinstance(r, float) and isinstance(sr, float):
            ratio = np.array([[r] * size, [sr] * size])
        elif isinstance(r, list) and isinstance(sr, list):
            ratio = np.array([r, sr])
        else:
            raise ValueError(f"Initial ratio is unsupported.")
        ar40a = calc.arr.mul(ar36a, ratio)
        ar40r = calc.arr.sub(ar40, ar40k, ar40a)
        ar40rar39k: list = calc.arr.div(ar40r, ar39k)
    except (IndexError, AttributeError, ValueError):
        raise ValueError(f"Check tables of corrected values and degas values.")
    else:
        return ar40rar39k


def get_plateau_results(smp: Sample, sequence: list, ar40rar39k: list = None,
                        ar39k_percentage: list = None, **kwargs):
    """
    Get initial ratio re-corrected plateau results
    Parameters
    ----------
    smp : sample instance
    sequence : data slice index
    ar40rar39k :
    ar39k_percentage : Ar39K released

    Returns
    -------
    three items tuple, result dict, age, and plot data, results keys = [
        'F', 'sF', 'Num', 'MSWD', 'Chisq', 'Pvalue',
        'age', 's1', 's2', 's3', 'Ar39', 'rs'
    ]
    """
    plateau_res_keys = [
        'F', 'sF', 'Num', 'MSWD', 'Chisq', 'Pvalue', 'age', 's1', 's2', 's3', 'Ar39',
        'rs',  # 'rs' means relative error of the total sum
    ]
    plateau_res = dict(zip(plateau_res_keys, [np.nan for i in plateau_res_keys]))

    def _get_partial(points, *args):
        return [arg[min(points): max(points) + 1] for arg in args]

    if len(sequence) == 0:
        return plateau_res, [], []
    if ar40rar39k is None:
        ar40rar39k = smp.ApparentAgeValues[0:2]
    if ar39k_percentage is None:
        ar39k_percentage = smp.ApparentAgeValues[7]

    if str(smp.Info.sample.type).lower() == "unknown":
        handle = basic.calc_age
    elif str(smp.Info.sample.type).lower() == "standard":
        handle = basic.calc_j
    elif str(smp.Info.sample.type).lower() == "air":
        handle = basic.calc_mdf
    else:
        raise TypeError(f"Sample type is not supported: {smp.Info.sample.type}")

    age = handle(ar40rar39k, smp=smp)[0:2]
    plot_data = calc.spectra.get_data(*age, ar39k_percentage, indices=sequence, **kwargs)
    f_values = _get_partial(sequence, *ar40rar39k)
    age = _get_partial(sequence, *age)
    sum_ar39k = sum(_get_partial(sequence, ar39k_percentage)[0])
    wmf = calc.arr.wtd_mean(*f_values)
    wmage = handle(wmf[0:2], smp=smp)

    plateau_res.update(dict(zip(plateau_res_keys, [*wmf, *wmage, sum_ar39k, np.nan])))
    return plateau_res, age, plot_data


def get_wma_results(sample: Sample, sequence: list, ar40rar39k: list = None, ar39k_percentage: list = None):
    """
    Get initial ratio re-corrected plateau results
    Parameters
    ----------
    sample : sample instance
    sequence : data slice index
    ar40rar39k :
    ar39k_percentage :

    Returns
    -------
    three itmes tuple, result dict, age, and plot data, results keys = [
        'F', 'sF', 'Num', 'MSWD', 'Chisq', 'Pvalue',
        'age', 's1', 's2', 's3', 'Ar39', 'rs'
    ]
    """
    if ar40rar39k is None:
        ar40rar39k = sample.ApparentAgeValues[0:2]
    if ar39k_percentage is None:
        ar39k_percentage = sample.ApparentAgeValues[7]

    spectra_res = initial.SPECTRA_RES.copy()
    # spectra_res = initial.SPECTRA_RES

    def _get_partial(points, *args):
        return [arg[min(points): max(points) + 1] for arg in args]

    if len(sequence) > 0:
        sum_ar39k = sum(_get_partial(sequence, ar39k_percentage)[0])
        fs = _get_partial(sequence, ar40rar39k[0])[0]
        sfs = _get_partial(sequence, ar40rar39k[1])[0]

        wmf, swmf, num, mswd, chisq, p = calc.arr.wtd_mean(fs, sfs)
        age, s1, s2, s3 = basic.calc_age([wmf, swmf], smp=sample)

        spectra_res.update({
            'age': age, 's1': s1, 's2': s2, 's3': s3, 'Num': num, 'MSWD': mswd, 'Chisq': chisq, 'Pvalue': p,
            'F': wmf, 'sF': swmf, 'Ar39': sum_ar39k
        })
    return spectra_res


def recalc_j_plateaus(sample: Sample, **kwargs):

    ar40rar39k = sample.ApparentAgeValues[0:2]
    j = sample.ApparentAgeValues[2:4]

    try:
        set1_res, _, set1_data = get_plateau_results(sample, sample.SelectedSequence1, ar40rar39k=ar40rar39k)
    except ValueError:
        pass
    else:
        sample.Info.results.age_plateau.update({0: set1_res})
        sample.AgeSpectraPlot.set1.data = calc.arr.transpose(set1_data)
        sample.AgeSpectraPlot.text1.text = ""  # 注意和js的配合，js那边根据text是否为空判断是否重新生成文字

    try:
        set2_res, _, set2_data = get_plateau_results(sample, sample.SelectedSequence2, ar40rar39k=ar40rar39k)
    except ValueError:
        pass
    else:
        sample.Info.results.age_plateau.update({1: set2_res})
        sample.AgeSpectraPlot.set2.data = calc.arr.transpose(set2_data)
        sample.AgeSpectraPlot.text2.text = ""  # 注意和js的配合，js那边根据text是否为空判断是否重新生成文字


# =======================
# Age Distribution Plot
# =======================
def recalc_agedistribution(sample: Sample, **kwargs):
    for i in range(2):
        try:
            # Age bars
            sample.AgeDistributionPlot.set3.data = calc.arr.remove(sample.ApparentAgeValues[2:4], (None, np.nan))
            # Set histogram data
            s = getattr(sample.AgeDistributionPlot.set1, 'bin_start', None)
            w = getattr(sample.AgeDistributionPlot.set1, 'bin_width', None)
            c = getattr(sample.AgeDistributionPlot.set1, 'bin_count', None)
            r = getattr(sample.AgeDistributionPlot.set1, 'bin_rule', None)
            # print(f's = {s}, r = {r}, w = {w}, c = {c}')
            histogram_data = calc.histogram.get_data(sample.ApparentAgeValues[2], s=s, r=r, w=w, c=c)
            sample.AgeDistributionPlot.set1.data = [histogram_data[1], histogram_data[0], histogram_data[2]]  # [half_bins, counts]
            setattr(sample.AgeDistributionPlot.set1, 'bin_start', histogram_data[3])
            setattr(sample.AgeDistributionPlot.set1, 'bin_rule', histogram_data[4])
            setattr(sample.AgeDistributionPlot.set1, 'bin_width', histogram_data[5])
            setattr(sample.AgeDistributionPlot.set1, 'bin_count', histogram_data[6])
            h = getattr(sample.AgeDistributionPlot.set2, 'band_width', None)
            k = getattr(sample.AgeDistributionPlot.set2, 'band_kernel', 'normal')
            t = getattr(sample.AgeDistributionPlot.set2, 'band_extend', False)
            a = getattr(sample.AgeDistributionPlot.set2, 'auto_width', 'Scott')
            n = getattr(sample.AgeDistributionPlot.set2, 'band_points', 1000)
            # print(f'h = {h}, k = {k}, a = {a}, n = {n}, extend = {t}')
            kda_data = calc.histogram.get_kde(
                sample.ApparentAgeValues[2], h=h, k=k, n=n, a=a,
                s=float(getattr(sample.AgeDistributionPlot.xaxis, 'min')) if t else histogram_data[3],
                e=float(getattr(sample.AgeDistributionPlot.xaxis, 'max')) if t else histogram_data[7],
            )
            sample.AgeDistributionPlot.set2.data = kda_data[0]  # [values, kda]
            setattr(sample.AgeDistributionPlot.set2, 'band_width', kda_data[1])
            setattr(sample.AgeDistributionPlot.set2, 'band_kernel', kda_data[2])
            setattr(sample.AgeDistributionPlot.set2, 'auto_width', kda_data[3])
            # sorted_data = [i[0] for i in sorted(zipped_data, key=lambda x: x[1])]
            text = f'n = {len(sample.ApparentAgeValues[2])}'
            peaks = find_peaks(kda_data[0][1])
            for index, peak in enumerate(peaks[0].tolist()):
                text = text + f'\nPeak {index}: {kda_data[0][0][peak]:.2f}'
            setattr(sample.AgeDistributionPlot.text1, 'text', text)
        except AttributeError:
            print(traceback.format_exc())
            initial.re_set_smp(sample)
            continue
        except (Exception, BaseException):
            print(traceback.format_exc())
            sample.AgeDistributionPlot.data = [[], []]
            sample.AgeDistributionPlot.set1.data = [[], []]
            sample.AgeDistributionPlot.set2.data = [[], []]
        break


